@@ -1,5 +1,5 @@
 /******************************************************************************
-*******************************************************************************
+******************************************************************************i
 **
 **  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
 **
@@ -34,6 +34,8 @@
 
 #include "copyright.cf"
 
+extern int quorate;
+int no_manager_opt=0;
 static int exit_now=0;
 static unsigned int flags=0;
 static sigset_t signal_mask;
@@ -74,9 +76,11 @@ int main(int argc, char *argv[]){
     free(msg);
   }
 
-  if(start_cluster_monitor_thread()){
-    log_err("Unable to create thread.\n");
-    exit(EXIT_FAILURE);
+  if (!no_manager_opt){
+    if(start_cluster_monitor_thread()){
+      log_err("Unable to create thread.\n");
+      exit(EXIT_FAILURE);
+    }
   }
 
   memset(&addr, 0, sizeof(struct sockaddr_storage));
@@ -177,8 +181,11 @@ int main(int argc, char *argv[]){
   if (sfds[2] >= 0) 
     FD_SET(sfds[2], &rset);
 
+  log_dbg("Sending SIGTERM to parent\n");
+  kill(getppid(), SIGTERM);
+
   while(1){
-    int len = addr_size;
+    unsigned int len = addr_size;
 
     process_signals();
     
@@ -266,6 +273,7 @@ static void print_usage(FILE *stream){
 	  " -t <ttl>      Multicast threshold (aka Time to Live) value.\n"
 	  " -P [bcf]:#    Specify various port numbers.\n"
 	  " -V            Print version information.\n"
+	  " -X            No cluster manager, just read local cluster.conf.\n"
 	  );
   EXIT("print_usage");
 }
@@ -324,12 +332,12 @@ static char *parse_cli_args(int argc, char *argv[]){
 
   memset(buff, 0, buff_size);
 
-  while((c = getopt(argc, argv, "46Icdf:hlm:nP:t:sV")) != -1){
+  while((c = getopt(argc, argv, "46Icdf:hlm:nP:t:sVX")) != -1){
     switch(c){
     case '4':
       if(IPv6 == 1){
 	fprintf(stderr,
-		"Setting protocol to IPv4 conflicts with previous protocol choice.\n");
+		"Setting protocol to IPv4 conflicts with multicast address.\n");
 	error = -EINVAL;
 	goto fail;
       }
@@ -356,12 +364,12 @@ static char *parse_cli_args(int argc, char *argv[]){
       use_local = 0;
       break;
     case 'c':
-      fprintf(stderr, "The '-c' option is depricated.\n"
+      fprintf(stderr, "The '-c' option is deprecated.\n"
 	      "Try '-h' for help.\n");
       error = -EINVAL;
       goto fail;
     case 'd':  /* might be usable for upgrade */
-      fprintf(stderr, "The '-d' option is depricated.\n"
+      fprintf(stderr, "The '-d' option is deprecated.\n"
 	      "Try '-h' for help.\n");
       error = -EINVAL;
       goto fail;
@@ -375,7 +383,7 @@ static char *parse_cli_args(int argc, char *argv[]){
       print_usage(stdout);
       exit(EXIT_SUCCESS);
     case 'l':
-      fprintf(stderr, "The '-l' option is depricated.\n"
+      fprintf(stderr, "The '-l' option is deprecated.\n"
 	      "Try '-h' for help.\n");
       error = -EINVAL;
       goto fail;
@@ -462,7 +470,7 @@ static char *parse_cli_args(int argc, char *argv[]){
       }
       break;
     case 's':
-      fprintf(stderr, "The '-s' option is depricated.\n"
+      fprintf(stderr, "The '-s' option is deprecated.\n"
 	      "Try '-h' for help.\n");
       error = -EINVAL;
       goto fail;
@@ -473,6 +481,10 @@ static char *parse_cli_args(int argc, char *argv[]){
       printf("%s %s (built %s %s)\n", argv[0], CCS_RELEASE_NAME, __DATE__, __TIME__);
       printf("%s\n", REDHAT_COPYRIGHT);
       exit(EXIT_SUCCESS);
+    case 'X':
+      no_manager_opt = 1;
+      quorate = 1;
+      break;
     default:
       print_usage(stderr);
       error = -EINVAL;
@@ -519,6 +531,10 @@ static int check_cluster_conf(void){
     xmlFreeDoc(doc);
   } else {
     /* no cluster.conf file.  This is fine, just need to get it from the network */
+    if(no_manager_opt){
+      log_err("\nNo local cluster.conf found: %s\n", config_file_location);
+      return -1;
+    }
   }
 
   return 0;
@@ -715,10 +731,6 @@ static void daemonize(void){
 	exit(EXIT_SUCCESS);
 
       switch(WEXITSTATUS(status)){
-      case EXIT_MAGMA_PLUGINS:
-	fprintf(stderr, "Failed to connect to cluster manager.\n"
-		"Hint: Magma plugins are not in the right spot.\n");
-	break;
       case EXIT_CLUSTER_FAIL:
 	fprintf(stderr, "Failed to connect to cluster manager.\n");
 	break;
@@ -745,9 +757,8 @@ static void daemonize(void){
       exit(EXIT_LOCKFILE);
     }
 
-    /* hold off on shutting down parent.  Let cluster_communicator do it **
-    ** after if figures out if it can use magma or not.................. **
-    **    kill(getppid(), SIGTERM); */
+    /* Make the parent stop waiting */
+    /* kill(getppid(), SIGTERM); */
   }
 
   signal(SIGINT, &sig_handler);
