@@ -22,6 +22,7 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <netdb.h>
+#include <time.h>
 #include <libxml/parser.h>
 #include <libxml/tree.h>
 #include <libxml/xpath.h>
@@ -33,19 +34,40 @@
 #include "misc.h"
 #include "globals.h"
 
+/* Default descriptor expiration time, in seconds */
+#ifndef DEFAULT_EXPIRE
+#define DEFAULT_EXPIRE 30
+#endif
+
+/* Maximum open connection count */
+#ifndef MAX_OPEN_CONNECTIONS
+#define MAX_OPEN_CONNECTIONS 30
+#endif
+
+/* Conversion from descriptor to ocs index */
+#ifdef dindex
+#undef dindex
+#endif
+#define dindex(x) ((x) % MAX_OPEN_CONNECTIONS)
+
+static inline void _cleanup_descriptor(int desc);
+
+extern int no_manager_opt;
 
 typedef struct open_connection_s {
   char *oc_cwp;
   char *oc_query;
-  int oc_index;
   open_doc_t *oc_odoc;
   xmlXPathContextPtr oc_ctx;
+  int oc_index;
+  int oc_desc;
+  time_t oc_expire;
 } open_connection_t;
 
 /* ATTENTION: need to lock on this if we start forking the daemon **
 **  Also would need to create a shared memory area for open cnx's */
-#define MAX_OPEN_CONNECTIONS 10
 static open_connection_t **ocs = NULL;
+static int _descbase = 0;
 
 static int _update_config(char *location){
   int error = 0;
@@ -148,8 +170,8 @@ static int broadcast_for_doc(char *cluster_name, int blocking){
   struct sockaddr_storage addr, recv_addr;
   struct sockaddr_in *addr4 = (struct sockaddr_in *)&addr;
   struct sockaddr_in6 *addr6 = (struct sockaddr_in6 *)&addr;
+  unsigned int len = sizeof(struct sockaddr_storage);
   int addr_size = 0;
-  int len=sizeof(struct sockaddr_storage);
   comm_header_t *ch = NULL;
   char *bdoc = NULL;
   fd_set rset;
@@ -270,6 +292,7 @@ static int broadcast_for_doc(char *cluster_name, int blocking){
     ch->comm_type = COMM_BROADCAST;
 
     log_dbg("Sending broadcast.\n");
+    swab_header(ch);
 
     if(sendto(sfd, (char *)ch, sizeof(comm_header_t), 0,
 	      (struct sockaddr *)&addr, addr_size) < 0){
@@ -304,6 +327,7 @@ static int broadcast_for_doc(char *cluster_name, int blocking){
 	error = 0;
 	recvfrom(sfd, (char *)ch, sizeof(comm_header_t), MSG_PEEK,
 		 (struct sockaddr *)&recv_addr, (socklen_t *)&len);
+	swab_header(ch);
 	if(!ch->comm_payload_size || ch->comm_error){
 	  /* clean out this reply by not using MSG_PEEK */
 	  recvfrom(sfd, (char *)ch, sizeof(comm_header_t), 0,
@@ -475,7 +499,7 @@ static int process_connect(comm_header_t *ch, char *cluster_name){
   int i=0, error = 0;
   int bcast_needed = 0;
   char *tmp_name = NULL;
-  
+  time_t now;
 
   ENTER("process_connect");
 
@@ -536,7 +560,8 @@ static int process_connect(comm_header_t *ch, char *cluster_name){
 	      tmp_name, error);
       /* We must check with the others to make sure this is valid. */
     }
-    bcast_needed = 1;
+    if (!no_manager_opt)
+      bcast_needed = 1;
     error = 0;
   } else {
     tmp_name = get_cluster_name(master_doc->od_doc);
@@ -595,12 +620,23 @@ static int process_connect(comm_header_t *ch, char *cluster_name){
     }
   }
 
+  /* Locate the connection descriptor */
+  now = time(NULL); 
   for(i=0; i < MAX_OPEN_CONNECTIONS; i++){
-    if(!ocs[i]){
-      break;
+    if (!ocs[i])
+      continue;
+    if (now >= ocs[i]->oc_expire) {
+      log_dbg("Recycling connection descriptor %d: Expired\n",
+	      ocs[i]->oc_desc );
+      _cleanup_descriptor(i);
     }
   }
 
+  for(i=0; i < MAX_OPEN_CONNECTIONS; i++){
+    if(!ocs[i])
+      break;
+  }
+
   if(i >= MAX_OPEN_CONNECTIONS){
     error = -EAGAIN;
     goto fail;
@@ -617,6 +653,19 @@ static int process_connect(comm_header_t *ch, char *cluster_name){
   master_doc->od_refs++;
   ocs[i]->oc_odoc = master_doc;
   ocs[i]->oc_ctx = xmlXPathNewContext(ocs[i]->oc_odoc->od_doc);
+  ocs[i]->oc_expire = now + DEFAULT_EXPIRE;
+
+  /* using error as a temp var */
+  error = i + _descbase++ * MAX_OPEN_CONNECTIONS;
+  if (error > INT_MAX || error < 0) {
+    error = i;
+    _descbase = 0;
+  }
+  ocs[i]->oc_desc = error;
+ 
+  /* reset error */
+  error = 0;
+
   if(!ocs[i]->oc_ctx){
     ocs[i]->oc_odoc->od_refs--;
     free(ocs[i]);
@@ -638,13 +687,47 @@ static int process_connect(comm_header_t *ch, char *cluster_name){
   if(error){
     ch->comm_error = error;
   } else {
-    ch->comm_desc = i;
+    ch->comm_desc = ocs[i]->oc_desc;
   }
   EXIT("process_connect");
   return error;
 }
 
 
+static inline void
+_cleanup_descriptor(int desc)
+{
+  open_doc_t *tmp_odoc;
+
+  if(ocs[desc]->oc_ctx){
+    xmlXPathFreeContext(ocs[desc]->oc_ctx);
+  }
+  if(ocs[desc]->oc_cwp){
+    free(ocs[desc]->oc_cwp);
+  }
+  if(ocs[desc]->oc_query){
+    free(ocs[desc]->oc_query);
+  }
+  tmp_odoc = ocs[desc]->oc_odoc;
+  if(tmp_odoc->od_refs < 1){
+    log_err("Number of references on an open doc should never be < 1.\n");
+    log_err("This is a fatal error.  Exiting...\n");
+    exit(EXIT_FAILURE);
+  }
+  if(tmp_odoc != master_doc && tmp_odoc->od_refs == 1){
+    log_dbg("No more references on version %d of config file, freeing...\n",
+	      get_doc_version(tmp_odoc->od_doc));
+    xmlFreeDoc(tmp_odoc->od_doc);
+    free(tmp_odoc);
+  } else {
+    tmp_odoc->od_refs--;
+  }
+
+  free(ocs[desc]);
+  ocs[desc] = NULL;
+}
+
+
 /**
  * process_disconnect: close an open session
  * @afd: accepted socket connection
@@ -655,53 +738,28 @@ static int process_connect(comm_header_t *ch, char *cluster_name){
  * Returns: 0 on success, < 0 on error
  */
 static int process_disconnect(comm_header_t *ch){
-  open_doc_t *tmp_odoc;
-  int desc = ch->comm_desc;
+  int desc = dindex(ch->comm_desc);
   int error=0;
   ENTER("process_disconnect");
 
   ch->comm_payload_size = 0;
 
-  if(desc < 0 || desc >= MAX_OPEN_CONNECTIONS){
+  if(desc < 0){
     log_err("Invalid descriptor specified (%d).\n", desc);
     log_err("Someone may be attempting something evil.\n");
     error = -EBADR;
     goto fail;
   }
 
-  if(!ocs || !ocs[desc]){
+  if(!ocs || !ocs[desc] || (ocs[desc]->oc_desc != ch->comm_desc)){
     /* send failure to requestor ? */
-    log_err("Attempt to close an unopened CCS descriptor (%d).\n", desc);
+    log_err("Attempt to close an unopened CCS descriptor (%d).\n",
+	    ch->comm_desc);
 
     error = -EBADR;
     goto fail;
   } else {
-    if(ocs[desc]->oc_ctx){
-      xmlXPathFreeContext(ocs[desc]->oc_ctx);
-    }
-    if(ocs[desc]->oc_cwp){
-      free(ocs[desc]->oc_cwp);
-    }
-    if(ocs[desc]->oc_query){
-      free(ocs[desc]->oc_query);
-    }
-    tmp_odoc = ocs[desc]->oc_odoc;
-    if(tmp_odoc->od_refs < 1){
-      log_err("Number of references on an open doc should never be < 1.\n");
-      log_err("This is a fatal error.  Exiting...\n");
-      exit(EXIT_FAILURE);
-    }
-    if(tmp_odoc != master_doc && tmp_odoc->od_refs == 1){
-      log_dbg("No more references on version %d of config file, freeing...\n",
-	      get_doc_version(tmp_odoc->od_doc));
-      xmlFreeDoc(tmp_odoc->od_doc);
-      free(tmp_odoc);
-    } else {
-      tmp_odoc->od_refs--;
-    }
-
-    free(ocs[desc]);
-    ocs[desc] = NULL;
+    _cleanup_descriptor(desc);
   }
 
  fail:
@@ -730,7 +788,7 @@ static int process_disconnect(comm_header_t *ch){
  * Returns: -EXXX on error, 1 if restarting list, 0 otherwise
  */
 static int _process_get(comm_header_t *ch, char **payload){
-  int error = 0;
+  int error = 0, desc = dindex(ch->comm_desc);
   xmlXPathObjectPtr obj = NULL;
   char *query = NULL;
 
@@ -741,37 +799,37 @@ static int _process_get(comm_header_t *ch, char **payload){
     goto fail;
   }
 
-  if(ch->comm_desc < 0 || ch->comm_desc >= MAX_OPEN_CONNECTIONS){
+  if(ch->comm_desc < 0){
     log_err("Invalid descriptor specified (%d).\n", ch->comm_desc);
     log_err("Someone may be attempting something evil.\n");
     error = -EBADR;
     goto fail;
   }
 
-  if(!ocs || !ocs[ch->comm_desc]){
+  if(!ocs || !ocs[desc] || (ocs[desc]->oc_desc != ch->comm_desc)){
     log_err("process_get: Invalid connection descriptor received.\n");
     error = -EBADR;
     goto fail;
   }
 
-  if(ocs[ch->comm_desc]->oc_query && !strcmp(*payload,ocs[ch->comm_desc]->oc_query)){
-    ocs[ch->comm_desc]->oc_index++;
-    log_dbg("Index = %d\n",ocs[ch->comm_desc]->oc_index);
+  if(ocs[desc]->oc_query && !strcmp(*payload,ocs[desc]->oc_query)){
+    ocs[desc]->oc_index++;
+    log_dbg("Index = %d\n",ocs[desc]->oc_index);
     log_dbg(" Query = %s\n", *payload);
   } else {
     log_dbg("Index reset (new query).\n");
     log_dbg(" Query = %s\n", *payload);
-    ocs[ch->comm_desc]->oc_index = 0;
-    if(ocs[ch->comm_desc]->oc_query){
-      free(ocs[ch->comm_desc]->oc_query);
+    ocs[desc]->oc_index = 0;
+    if(ocs[desc]->oc_query){
+      free(ocs[desc]->oc_query);
     }
-    ocs[ch->comm_desc]->oc_query = (char *)strdup(*payload);
+    ocs[desc]->oc_query = (char *)strdup(*payload);
   }
 
   /* ATTENTION -- should path expansion go before index inc ? */
   if(((ch->comm_payload_size > 1) &&
       ((*payload)[0] == '/')) ||
-     !ocs[ch->comm_desc]->oc_cwp){
+     !ocs[desc]->oc_cwp){
     log_dbg("Query involves absolute path or cwp is not set.\n");
     query = (char *)strdup(*payload);
     if(!query){
@@ -781,15 +839,18 @@ static int _process_get(comm_header_t *ch, char **payload){
   } else {
     /* +2 because of NULL and '/' character */
     log_dbg("Query involves relative path.\n");
-    query = malloc(strlen(*payload)+strlen(ocs[ch->comm_desc]->oc_cwp)+2);
+    query = malloc(strlen(*payload)+strlen(ocs[desc]->oc_cwp)+2);
     if(!query){
       error = -ENOMEM;
       goto fail;
     }
-    sprintf(query, "%s/%s", ocs[ch->comm_desc]->oc_cwp, *payload);
+    sprintf(query, "%s/%s", ocs[desc]->oc_cwp, *payload);
   }
 
-  obj = xmlXPathEvalExpression(query, ocs[ch->comm_desc]->oc_ctx);
+  /* Bump expiration time */
+  ocs[desc]->oc_expire = time(NULL) + DEFAULT_EXPIRE;
+
+  obj = xmlXPathEvalExpression((xmlChar *)query, ocs[desc]->oc_ctx);
   if(obj){
     log_dbg("Obj type  = %d (%s)\n", obj->type, (obj->type == 1)?"XPATH_NODESET":"");
     log_dbg("Number of matches: %d\n", (obj->nodesetval)?obj->nodesetval->nodeNr:0);
@@ -798,20 +859,20 @@ static int _process_get(comm_header_t *ch, char **payload){
       int size=0;
       int nnv=0; /* name 'n' value */
 
-      if(ocs[ch->comm_desc]->oc_index >= obj->nodesetval->nodeNr){
-	ocs[ch->comm_desc]->oc_index = 0;
+      if(ocs[desc]->oc_index >= obj->nodesetval->nodeNr){
+	ocs[desc]->oc_index = 0;
 	error = 1;
 	log_dbg("Index reset to zero (end of list).\n");
       }
 	  
-      node = obj->nodesetval->nodeTab[ocs[ch->comm_desc]->oc_index];
+      node = obj->nodesetval->nodeTab[ocs[desc]->oc_index];
 	
       log_dbg("Node (%s) type = %d (%s)\n", node->name, node->type,
 	      (node->type == 1)? "XML_ELEMENT_NODE":
 	      (node->type == 2)? "XML_ATTRIBUTE_NODE":"");
 
       if(!node || !node->children ||
-	 !node->children->content || !strlen(node->children->content)){
+         !node->children->content || !strlen((char *)node->children->content)){
 	log_dbg("No content found.\n");
 	error = -ENODATA;
 	goto fail;
@@ -823,10 +884,11 @@ static int _process_get(comm_header_t *ch, char **payload){
 	 ((node->type == XML_ELEMENT_NODE) && strstr(query, "child::*"))){
 	/* add on the trailing NULL and the '=' separator for a list of attrs
 	   or an element node + CDATA*/
-	size = strlen(node->children->content)+strlen(node->name)+2;
+	size = strlen((char *)node->children->content) +
+	  strlen((char *)node->name)+2;
 	nnv= 1;
       } else {
-	size = strlen(node->children->content)+1;
+	size = strlen((char *)node->children->content) + 1;
       }
 
       if(size <= ch->comm_payload_size){  /* do we already have enough space? */
@@ -893,8 +955,8 @@ static int process_get_list(comm_header_t *ch, char **payload){
   error = _process_get(ch, payload);
   if(error){
     ch->comm_payload_size = 0;
-    if(ocs && ocs[ch->comm_desc])
-      ocs[ch->comm_desc]->oc_index = -1;
+    if(ocs && ocs[dindex(ch->comm_desc)])
+      ocs[dindex(ch->comm_desc)]->oc_index = -1;
   }
 
   EXIT("process_get_list");
@@ -903,6 +965,7 @@ static int process_get_list(comm_header_t *ch, char **payload){
 
 static int process_set(comm_header_t *ch, char *payload){
   int error = 0;
+  int desc = dindex(ch->comm_desc);
 
   ENTER("process_set");
   if(!ch->comm_payload_size){
@@ -911,14 +974,14 @@ static int process_set(comm_header_t *ch, char *payload){
     goto fail;
   }
 
-  if(ch->comm_desc < 0 || ch->comm_desc >= MAX_OPEN_CONNECTIONS){
+  if(ch->comm_desc < 0){
     log_err("Invalid descriptor specified (%d).\n", ch->comm_desc);
     log_err("Someone may be attempting something evil.\n");
     error = -EBADR;
     goto fail;
   }
 
-  if(!ocs || !ocs[ch->comm_desc]){
+  if(!ocs || !ocs[desc] || (ocs[desc]->oc_desc != ch->comm_desc)){
     log_err("process_set: Invalid connection descriptor received.\n");
     error = -EBADR;
     goto fail;
@@ -938,7 +1001,7 @@ static int process_set(comm_header_t *ch, char *payload){
 
 
 static int process_get_state(comm_header_t *ch, char **payload){
-  int error = 0;
+  int error = 0, desc = dindex(ch->comm_desc);
   char *load = NULL;
 
   ENTER("process_get_state");
@@ -948,41 +1011,41 @@ static int process_get_state(comm_header_t *ch, char **payload){
     goto fail;
   }
 
-  if(ch->comm_desc < 0 || ch->comm_desc >= MAX_OPEN_CONNECTIONS){
+  if(ch->comm_desc < 0){
     log_err("Invalid descriptor specified (%d).\n", ch->comm_desc);
     log_err("Someone may be attempting something evil.\n");
     error = -EBADR;
     goto fail;
   }
 
-  if(!ocs || !ocs[ch->comm_desc]){
+  if(!ocs || !ocs[desc] || (ocs[desc]->oc_desc != ch->comm_desc)){
     log_err("process_get_state: Invalid connection descriptor received.\n");
     error = -EBADR;
     goto fail;
   }
 
-  if(ocs[ch->comm_desc]->oc_cwp && ocs[ch->comm_desc]->oc_query){
-    int size = strlen(ocs[ch->comm_desc]->oc_cwp) +
-      strlen(ocs[ch->comm_desc]->oc_query) + 2;
+  if(ocs[desc]->oc_cwp && ocs[desc]->oc_query){
+    int size = strlen(ocs[desc]->oc_cwp) +
+      strlen(ocs[desc]->oc_query) + 2;
     log_dbg("Both cwp and query are set.\n");
     load = malloc(size);
     if(!load){
       error = -ENOMEM;
       goto fail;
     }
-    strcpy(load, ocs[ch->comm_desc]->oc_cwp);
-    strcpy(load+strlen(ocs[ch->comm_desc]->oc_cwp)+1, ocs[ch->comm_desc]->oc_query);
+    strcpy(load, ocs[desc]->oc_cwp);
+    strcpy(load+strlen(ocs[desc]->oc_cwp)+1, ocs[desc]->oc_query);
     ch->comm_payload_size = size;
-  } else if(ocs[ch->comm_desc]->oc_cwp){
+  } else if(ocs[desc]->oc_cwp){
     log_dbg("Only cwp is set.\n");
-    load = (char *)strdup(ocs[ch->comm_desc]->oc_cwp);
+    load = (char *)strdup(ocs[desc]->oc_cwp);
     if(!load){
       error = -ENOMEM;
       goto fail;
     }
     ch->comm_payload_size = strlen(load)+1;
-  } else if(ocs[ch->comm_desc]->oc_query){
-    int size = strlen(ocs[ch->comm_desc]->oc_query) + 2;
+  } else if(ocs[desc]->oc_query){
+    int size = strlen(ocs[desc]->oc_query) + 2;
     log_dbg("Only query is set.\n");
     load = malloc(size);
     if(!load){
@@ -990,10 +1053,11 @@ static int process_get_state(comm_header_t *ch, char **payload){
       goto fail;
     }
     memset(load, 0, size);
-    strcpy(load+1, ocs[ch->comm_desc]->oc_query);
+    strcpy(load+1, ocs[desc]->oc_query);
     ch->comm_payload_size = size;
   }
 
+  ocs[desc]->oc_expire = time(NULL) + DEFAULT_EXPIRE;
   *payload = load;
 
  fail:
@@ -1008,7 +1072,7 @@ static int process_get_state(comm_header_t *ch, char **payload){
 
 
 static int process_set_state(comm_header_t *ch, char *payload){
-  int error = 0;
+  int error = 0, desc = dindex(ch->comm_desc);
 
   ENTER("process_set_state");
   if(!ch->comm_payload_size){
@@ -1024,23 +1088,24 @@ static int process_set_state(comm_header_t *ch, char *payload){
     goto fail;
   }
 
-  if(!ocs || !ocs[ch->comm_desc]){
+  if(!ocs || !ocs[desc] || (ocs[desc]->oc_desc != ch->comm_desc)){
     log_err("process_set_state: Invalid connection descriptor received.\n");
     error = -EBADR;
     goto fail;
   }
 
-  if(ocs[ch->comm_desc]->oc_cwp){
-    free(ocs[ch->comm_desc]->oc_cwp);
-    ocs[ch->comm_desc]->oc_cwp = NULL;
+  if(ocs[desc]->oc_cwp){
+    free(ocs[desc]->oc_cwp);
+    ocs[desc]->oc_cwp = NULL;
   }
 
-  if((ch->comm_flags & COMM_SET_STATE_RESET_QUERY) && ocs[ch->comm_desc]->oc_query){
-    free(ocs[ch->comm_desc]->oc_query);
-    ocs[ch->comm_desc]->oc_query = NULL;
+  if((ch->comm_flags & COMM_SET_STATE_RESET_QUERY) && ocs[desc]->oc_query){
+    free(ocs[desc]->oc_query);
+    ocs[desc]->oc_query = NULL;
   }
 
-  ocs[ch->comm_desc]->oc_cwp = (char *)strdup(payload);
+  ocs[desc]->oc_expire = time(NULL) + DEFAULT_EXPIRE;
+  ocs[desc]->oc_cwp = (char *)strdup(payload);
 
  fail:
   ch->comm_payload_size = 0;
@@ -1171,6 +1236,7 @@ int process_request(int afd){
   error = write(afd, ch, sizeof(comm_header_t)+ch->comm_payload_size);
   if(error < 0){
     log_sys_err("Unable to write package back to sender");
+    return error; 
     goto fail;
   } else if(error < (sizeof(comm_header_t)+ch->comm_payload_size)){
     log_err("Unable to write complete package.\n");
@@ -1198,10 +1264,11 @@ int process_request(int afd){
 int process_broadcast(int sfd){
   int error = 0;
   comm_header_t *ch = NULL;
-  char *payload = NULL;
+  xmlChar *payload = NULL;
   char *buffer = NULL;
   struct sockaddr_storage addr;
-  int len = sizeof(struct sockaddr_storage);  /* value/result for recvfrom */
+  unsigned int len = sizeof(struct sockaddr_storage);  /* value/result for recvfrom */
+  int sendlen;
   int discard = 0;
 
   ENTER("process_broadcast");
@@ -1220,12 +1287,13 @@ int process_broadcast(int sfd){
     error = -errno;
     goto fail;
   }
+  swab_header(ch);
 
   if(ch->comm_type != COMM_BROADCAST){
     /* Either someone is pinging this port, or there is an older version **
     ** of ccs trying to get bcast response.  Either way, we should not   **
     ** respond to them.................................................. */
-    log_dbg("Received invalid request on broadcast port.\n");
+    log_dbg("Received invalid request on broadcast port. %x\n",ch->comm_type);
     error = -EINVAL;
     goto fail;
   }
@@ -1264,7 +1332,7 @@ int process_broadcast(int sfd){
 
   /* allocates space for the payload */
   xmlDocDumpFormatMemory(master_doc->od_doc,
-			 (xmlChar **)&payload,
+			 &payload,
 			 &(ch->comm_payload_size),
 			 0);
   if(!ch->comm_payload_size){
@@ -1287,7 +1355,9 @@ int process_broadcast(int sfd){
   memcpy(buffer+sizeof(comm_header_t), payload, ch->comm_payload_size);
 
   log_dbg("Sending cluster.conf (version %d)...\n", get_doc_version(master_doc->od_doc));
-  if(sendto(sfd, buffer, ch->comm_payload_size + sizeof(comm_header_t), 0,
+  sendlen = ch->comm_payload_size + sizeof(comm_header_t);
+  swab_header(ch);
+  if(sendto(sfd, buffer, sendlen, 0,
 	    (struct sockaddr *)&addr, (socklen_t)len) < 0){
     log_sys_err("Sendto failed");
     error = -errno;
