@@ -2,7 +2,7 @@
 *******************************************************************************
 **
 **  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
+**  Copyright (C) 2004-2006 Red Hat, Inc.  All rights reserved.
 **
 **  This copyrighted material is made available to anyone wishing to use,
 **  modify, copy, or redistribute it subject to the terms and conditions
@@ -19,6 +19,7 @@
 #include <linux/completion.h>
 #include <linux/buffer_head.h>
 #include <asm/uaccess.h>
+#include <linux/lm_interface.h>
 
 #include "gfs.h"
 #include "dio.h"
@@ -1159,10 +1160,15 @@ gfs_glock_xmote_th(struct gfs_glock *gl, unsigned int state, int flags)
 			      gl->gl_state, state,
 			      lck_flags);
 
+	if (gfs_assert_withdraw(sdp, !(lck_ret & LM_OUT_ERROR)))
+		goto out;
+
 	if (lck_ret & LM_OUT_ASYNC)
 		gfs_assert_warn(sdp, lck_ret == LM_OUT_ASYNC);
 	else
 		xmote_bh(gl, lck_ret);
+ out:
+	return;
 }
 
 /**
@@ -1250,10 +1256,15 @@ gfs_glock_drop_th(struct gfs_glock *gl)
 
 	ret = gfs_lm_unlock(sdp, gl->gl_lock, gl->gl_state);
 
+	if (gfs_assert_withdraw(sdp, !(ret & LM_OUT_ERROR)))
+		goto out;
+
 	if (!ret)
 		drop_bh(gl, ret);
 	else
 		gfs_assert_warn(sdp, ret == LM_OUT_ASYNC);
+ out:
+	return;
 }
 
 /**
@@ -1607,6 +1618,9 @@ gfs_glock_dq(struct gfs_holder *gh)
 	struct gfs_glock *gl = gh->gh_gl;
 	struct gfs_sbd *sdp = gl->gl_sbd;
 	struct gfs_glock_operations *glops = gl->gl_ops;
+	struct list_head *pos;
+	struct gfs_holder *tmp_gh = NULL;
+	int count = 0;
 
 	atomic_inc(&gl->gl_sbd->sd_glock_dq_calls);
 
@@ -1617,8 +1631,14 @@ gfs_glock_dq(struct gfs_holder *gh)
 		set_bit(GLF_SYNC, &gl->gl_flags);
 
 	/* Don't cache glock; request demote to unlock at inter-node scope */
-	if (gh->gh_flags & GL_NOCACHE)
-		handle_callback(gl, LM_ST_UNLOCKED);
+	if (gh->gh_flags & GL_NOCACHE) {
+		list_for_each(pos, &gl->gl_holders) {
+			tmp_gh = list_entry(pos, struct gfs_holder, gh_list);
+			++count;
+		}
+		if (tmp_gh == gh && count == 1)
+			handle_callback(gl, LM_ST_UNLOCKED);
+	}
 
 	lock_on_glock(gl);
 
@@ -2117,6 +2137,7 @@ gfs_lvb_unhold(struct gfs_glock *gl)
 	glock_put(gl);
 }
 
+#if 0
 /**
  * gfs_lvb_sync - sync a LVB
  * @gl: The glock in question
@@ -2136,6 +2157,7 @@ gfs_lvb_sync(struct gfs_glock *gl)
 
 	unlock_on_glock(gl);
 }
+#endif
 
 /**
  * blocking_cb -
@@ -2185,9 +2207,9 @@ blocking_cb(struct gfs_sbd *sdp, struct lm_lockname *name, unsigned int state)
  */
 
 void
-gfs_glock_cb(lm_fsdata_t *fsdata, unsigned int type, void *data)
+gfs_glock_cb(void *fsdata, unsigned int type, void *data)
 {
-	struct gfs_sbd *sdp = (struct gfs_sbd *)fsdata;
+	struct gfs_sbd *sdp = fsdata;
 
 	atomic_inc(&sdp->sd_lm_callbacks);
 
