@@ -2,7 +2,7 @@
 *******************************************************************************
 **
 **  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
+**  Copyright (C) 2004-2006 Red Hat, Inc.  All rights reserved.
 **
 **  This copyrighted material is made available to anyone wishing to use,
 **  modify, copy, or redistribute it subject to the terms and conditions
@@ -512,7 +512,7 @@ struct gfs_glock {
 	struct gfs_holder *gl_req_gh; /* Holder for request being serviced */
 	gfs_glop_bh_t gl_req_bh;  /* The bottom half to execute */
 
-	lm_lock_t *gl_lock;       /* Lock module's private lock data */
+	void *gl_lock;            /* Lock module's private lock data */
 	char *gl_lvb;             /* Lock Value Block */
 	atomic_t gl_lvb_count;    /* LVB recursive usage (hold/unhold) count */
 
@@ -617,6 +617,7 @@ struct gfs_inode {
 
 	unsigned int i_greedy; /* The amount of time to be greedy */
 	unsigned long i_last_pfault; /* The time of the last page fault */
+	struct address_space_operations gfs_file_aops;
 };
 
 /*
@@ -817,12 +818,23 @@ struct gfs_trans {
 #define GFS_GLOCKD_DEFAULT (1)
 #define GFS_GLOCKD_MAX (32)
 
+#define GFS_QUOTA_DEFAULT      GFS_QUOTA_OFF
+#define GFS_QUOTA_OFF          0
+#define GFS_QUOTA_ACCOUNT      1
+#define GFS_QUOTA_ON           2
+
+#define GFS_DATA_DEFAULT       GFS_DATA_ORDERED
+#define GFS_DATA_WRITEBACK     1
+#define GFS_DATA_ORDERED       2
+
+
 struct gfs_args {
 	char ar_lockproto[GFS_LOCKNAME_LEN]; /* The name of the Lock Protocol */
 	char ar_locktable[GFS_LOCKNAME_LEN]; /* The name of the Lock Table */
 	char ar_hostdata[GFS_LOCKNAME_LEN]; /* The host specific data */
 
-        /*
+	int ar_spectator; /* Don't get a journal because we're always RO. */
+	/*
 	 * GFS can invoke some flock and disk caching optimizations if it is
 	 * not in a cluster, i.e. is a local filesystem.  The chosen lock
 	 * module tells GFS, at mount time, if it supports clustering.
@@ -1049,6 +1061,7 @@ struct gfs_sbd {
 
 	/* Scan for glocks and inodes to toss from memory */
 	struct task_struct *sd_scand_process; /* Scand places on reclaim list*/
+	struct task_struct *sd_glockd_process[GFS_GLOCKD_MAX];
 	unsigned int sd_glockd_num;    /* # of glockd procs to do reclaiming*/
 
 	/* Recover journal of a crashed node */
@@ -1191,6 +1204,10 @@ struct gfs_sbd {
 	atomic_t sd_ops_vm;
 
 	char sd_fsname[256];
+	char sd_table_name[256];
+	char sd_proto_name[256];
+
+	struct kobject sd_kobj;
 
 	/* Debugging crud */
 
