@@ -2,7 +2,7 @@
 *******************************************************************************
 **
 **  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
+**  Copyright (C) 2004-2006 Red Hat, Inc.  All rights reserved.
 **
 **  This copyrighted material is made available to anyone wishing to use,
 **  modify, copy, or redistribute it subject to the terms and conditions
@@ -19,8 +19,8 @@
 #include <linux/completion.h>
 #include <linux/buffer_head.h>
 #include <asm/uaccess.h>
-#include <linux/gfs_ioctl.h>
 
+#include "gfs_ioctl.h"
 #include "gfs.h"
 #include "bmap.h"
 #include "dio.h"
@@ -432,8 +432,8 @@ gi_get_tune(struct gfs_inode *ip,
 	    unsigned int size,
 	    unsigned int *count)
 {
-        struct gfs_tune *gt = &ip->i_sbd->sd_tune;
-        int error = -ENOBUFS;
+	struct gfs_tune *gt = &ip->i_sbd->sd_tune;
+	int error = -ENOBUFS;
 
 	if (gi->gi_argc != 1)
 		return -EINVAL;
@@ -479,6 +479,7 @@ gi_get_tune(struct gfs_inode *ip,
         gfs_printf("greedy_default %u\n", gt->gt_greedy_default);
         gfs_printf("greedy_quantum %u\n", gt->gt_greedy_quantum);
         gfs_printf("greedy_max %u\n", gt->gt_greedy_max);
+        gfs_printf("rgrp_try_threshold %u\n", gt->gt_rgrp_try_threshold);
 
         error = 0;
 
@@ -738,6 +739,11 @@ gi_set_tune(struct gfs_sbd *sdp, struct gfs_ioctl *gi)
 			return -EINVAL;
 		tune_set(gt_greedy_max, x);
 
+	} else if (strcmp(param, "rgrp_try_threshold") == 0) {
+		if (sscanf(value, "%u", &x) != 1)
+			return -EINVAL;
+		tune_set(gt_rgrp_try_threshold, x);
+
 	} else
 		return -EINVAL;
 
