@@ -1,29 +1,25 @@
-/******************************************************************************
-*******************************************************************************
-**
-**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
-**
-**  This copyrighted material is made available to anyone wishing to use,
-**  modify, copy, or redistribute it subject to the terms and conditions
-**  of the GNU General Public License v.2.
-**
-*******************************************************************************
-******************************************************************************/
+/*
+ *  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
+ *  Copyright (C) 2004-2006 Red Hat, Inc.  All rights reserved.
+ *
+ *  This copyrighted material is made available to anyone wishing to use,
+ *  modify, copy, or redistribute it subject to the terms and conditions
+ *  of the GNU General Public License v.2.
+ *
+ */
 
 #include <linux/sched.h>
 #include <linux/slab.h>
 #include <linux/smp_lock.h>
 #include <linux/spinlock.h>
-#include <asm/semaphore.h>
 #include <linux/completion.h>
 #include <linux/buffer_head.h>
 #include <linux/vmalloc.h>
 #include <linux/blkdev.h>
+#include <linux/kthread.h>
 
 #include "gfs.h"
 #include "daemon.h"
-#include "diaper.h"
 #include "glock.h"
 #include "glops.h"
 #include "inode.h"
@@ -37,58 +33,28 @@
 #include "recovery.h"
 #include "rgrp.h"
 #include "super.h"
+#include "sys.h"
 #include "unlinked.h"
 
-/**
- * gfs_read_super - Read in superblock
- * @sb: The VFS superblock
- * @data: Mount options
- * @silent: Don't complain if it's not a GFS filesystem
- *
- * Returns: errno
- *
- * After cross-linking Linux VFS incore superblock and our GFS incore superblock
- *   (filesystem instance structures) to one another, we:
- * -- Init some of our GFS incore superblock, including some temporary
- *       block-size values (enough to read on-disk superblock).
- * -- Set up some things in Linux VFS superblock.
- * -- Mount a lock module, init glock system (incl. glock reclaim daemons),
- *       and init some important inter-node locks (MOUNT, LIVE, SuperBlock).
- * -- Read-in the GFS on-disk superblock (1st time, to get enough info
- *       to do filesystem upgrade and journal replay, incl. journal index).
- * -- Upgrade on-disk filesystem format (rarely needed).
- * -- Replay journal(s) (always; replay *all* journals if we're first-to-mount).
- * -- Read-in on-disk superblock and journal index special file again (2nd time,
- *       assumed 100% valid now after journal replay).
- * -- Read-in info on other special (hidden) files (root inode, resource index,
- *       quota inode, license inode).
- * -- Start other daemons (journal/log recovery, log tail, quota updates, inode
- *       reclaim) for periodic maintenance.
- * 
- */
+#define DO 0
+#define UNDO 1
 
-static int
-fill_super(struct super_block *sb, void *data, int silent)
+extern struct dentry_operations gfs_dops;
+
+static struct gfs_sbd *init_sbd(struct super_block *sb)
 {
 	struct gfs_sbd *sdp;
-	struct gfs_holder mount_gh, sb_gh, ji_gh;
-	struct inode *inode;
-	int super = TRUE, jindex = TRUE;
 	unsigned int x;
-	int error;
 
 	sdp = vmalloc(sizeof(struct gfs_sbd));
-	if (!sdp) {
-		printk("GFS: can't alloc struct gfs_sbd\n");
-		error = -ENOMEM;
-		goto fail;
-	}
+	if (!sdp)
+		return NULL;
 
 	memset(sdp, 0, sizeof(struct gfs_sbd));
 
 	set_v2sdp(sb, sdp);
 	sdp->sd_vfs = sb;
-	gfs_diaper_register_sbd(sb->s_bdev, sdp);
+	gfs_tune_init(&sdp->sd_tune);
 
 	/*  Init rgrp variables  */
 
@@ -140,20 +106,12 @@ fill_super(struct super_block *sb, void *data, int silent)
 	spin_lock_init(&sdp->sd_ail_lock);
 	INIT_LIST_HEAD(&sdp->sd_recovery_bufs);
 
-	gfs_tune_init(&sdp->sd_tune);
+	return sdp;
+}
 
-	error = gfs_make_args((char *)data, &sdp->sd_args);
-	if (error) {
-		printk("GFS: can't parse mount arguments\n");
-		goto fail_vfree;
-	}
-
-	/*  Copy VFS mount flags  */
-
-	if (sb->s_flags & (MS_NOATIME | MS_NODIRATIME))
-		set_bit(SDF_NOATIME, &sdp->sd_flags);
-	if (sb->s_flags & MS_RDONLY)
-		set_bit(SDF_ROFS, &sdp->sd_flags);
+static void init_vfs(struct super_block *sb, unsigned noatime)
+{
+	struct gfs_sbd *sdp = sb->s_fs_info;
 
 	/*  Set up Linux Virtual (VFS) Super Block  */
 
@@ -169,28 +127,74 @@ fill_super(struct super_block *sb, void *data, int silent)
 	    lists), tell VFS */
 	if (sdp->sd_args.ar_posix_acls)
 		sb->s_flags |= MS_POSIXACL;
+}
 
-	/*  Set up the buffer cache and fill in some fake block size values
-	   to allow us to read-in the on-disk superblock.  */
+int init_names(struct gfs_sbd *sdp, int silent)
+{
+	struct gfs_sb *sb = NULL;
+	char *proto, *table;
+	int error = 0;
 
-	sdp->sd_sb.sb_bsize = sb_min_blocksize(sb, GFS_BASIC_BLOCK);
-	sdp->sd_sb.sb_bsize_shift = sb->s_blocksize_bits;
-	sdp->sd_fsb2bb_shift = sdp->sd_sb.sb_bsize_shift - GFS_BASIC_BLOCK_SHIFT;
-	sdp->sd_fsb2bb = 1 << sdp->sd_fsb2bb_shift;
+	proto = sdp->sd_args.ar_lockproto;
+	table = sdp->sd_args.ar_locktable;
 
-	if (sizeof(struct gfs_sb) > sdp->sd_sb.sb_bsize) {
-		printk("GFS: sizeof(struct gfs_sb) > sdp->sd_sb.sb_bsize\n"
-		       "GFS: %u > %u\n",
-		       (unsigned int)sizeof(struct gfs_sb), sdp->sd_sb.sb_bsize);
-		error = -EINVAL;
-		goto fail_vfree;
+	/*  Try to autodetect  */
+
+	if (!proto[0] || !table[0]) {
+		struct buffer_head *bh;
+
+		bh = sb_getblk(sdp->sd_vfs,
+			       GFS_SB_ADDR >> sdp->sd_fsb2bb_shift);
+		lock_buffer(bh);
+		clear_buffer_uptodate(bh);
+		clear_buffer_dirty(bh);
+		unlock_buffer(bh);
+		ll_rw_block(READ, 1, &bh);
+		wait_on_buffer(bh);
+
+		if (!buffer_uptodate(bh)) {
+			brelse(bh);
+			return -EIO;
+		}
+
+		sb = kmalloc(sizeof(struct gfs_sb), GFP_KERNEL);
+		if (!sb) {
+			brelse(bh);
+			return -ENOMEM;
+		}
+		gfs_sb_in(sb, bh->b_data); 
+		brelse(bh);
+
+		error = gfs_check_sb(sdp, sb, silent);
+		if (error)
+			goto out;
+
+		if (!proto[0])
+			proto = sb->sb_lockproto;
+		if (!table[0])
+			table = sb->sb_locktable;
 	}
 
-	/*  Mount an inter-node lock module, check for local optimizations */
+	if (!table[0])
+		table = sdp->sd_vfs->s_id;
 
-	error = gfs_lm_mount(sdp, silent);
-	if (error)
-		goto fail_vfree;
+	snprintf(sdp->sd_proto_name, 256, "%s", proto);
+	snprintf(sdp->sd_table_name, 256, "%s", table);
+
+ out:
+	kfree(sb);
+
+	return error;
+}
+
+static int init_locking(struct gfs_sbd *sdp, struct gfs_holder *mount_gh,
+						int undo)
+{
+	struct task_struct *p;
+	int error = 0;
+
+	if (undo)
+		goto fail_live;
 
 	if ((sdp->sd_lockstruct.ls_flags & LM_LSFLAG_LOCAL) &&
 	    !sdp->sd_args.ar_ignore_local_fs) {
@@ -206,37 +210,39 @@ fill_super(struct super_block *sb, void *data, int silent)
 
 	/*  Start up the scand thread  */
 
-	error = kernel_thread(gfs_scand, sdp, 0);
-	if (error < 0) {
+	p = kthread_run(gfs_scand, sdp, "gfs_scand");
+	error = IS_ERR(p);
+	if (error) {
 		printk("GFS: fsid=%s: can't start scand thread: %d\n",
 		       sdp->sd_fsname, error);
-		goto fail_lockproto;
+		return error;
 	}
-	wait_for_completion(&sdp->sd_thread_completion);
+	sdp->sd_scand_process = p;
 
 	/*  Start up the glockd thread  */
 
 	for (sdp->sd_glockd_num = 0;
 	     sdp->sd_glockd_num < sdp->sd_args.ar_num_glockd;
 	     sdp->sd_glockd_num++) {
-		error = kernel_thread(gfs_glockd, sdp, 0);
-		if (error < 0) {
+		p = kthread_run(gfs_glockd, sdp, "gfs_glockd");
+		error = IS_ERR(p);
+		if (error) {
 			printk("GFS: fsid=%s: can't start glockd thread: %d\n",
 			       sdp->sd_fsname, error);
-			goto fail_glockd;
+			goto fail;
 		}
-		wait_for_completion(&sdp->sd_thread_completion);
+		sdp->sd_glockd_process[sdp->sd_glockd_num] = p;
 	}
 
 	/*  Only one node may mount at a time */
 	error = gfs_glock_nq_num(sdp,
 				 GFS_MOUNT_LOCK, &gfs_nondisk_glops,
 				 LM_ST_EXCLUSIVE, LM_FLAG_NOEXP | GL_NOCACHE,
-				 &mount_gh);
+				 mount_gh);
 	if (error) {
 		printk("GFS: fsid=%s: can't acquire mount glock: %d\n",
 		       sdp->sd_fsname, error);
-		goto fail_glockd;
+		goto fail;
 	}
 
 	/*  Show that cluster is alive */
@@ -247,10 +253,37 @@ fill_super(struct super_block *sb, void *data, int silent)
 	if (error) {
 		printk("GFS: fsid=%s: can't acquire live glock: %d\n",
 		       sdp->sd_fsname, error);
-		goto fail_gunlock_mount;
+		goto fail_mount;
 	}
 
 	sdp->sd_live_gh.gh_owner = NULL;
+	return 0;
+
+fail_live:
+	gfs_glock_dq_uninit(&sdp->sd_live_gh);
+
+fail_mount:
+	gfs_glock_dq_uninit(mount_gh);
+
+fail:
+	while (sdp->sd_glockd_num--)
+		kthread_stop(sdp->sd_glockd_process[sdp->sd_glockd_num]);
+
+	kthread_stop(sdp->sd_scand_process);
+
+	return error;
+}
+
+static int init_sb(struct gfs_sbd *sdp, int silent, int undo)
+{
+	struct super_block *sb = sdp->sd_vfs;
+	struct gfs_holder sb_gh;
+	int error = 0;
+	int super = TRUE;
+
+	if (undo) {
+		return 0;
+	}
 
 	/*  Read the SuperBlock from disk, get enough info to enable us
 	    to read-in the journal index and replay all journals. */
@@ -262,14 +295,14 @@ fill_super(struct super_block *sb, void *data, int silent)
 	if (error) {
 		printk("GFS: fsid=%s: can't acquire superblock glock: %d\n",
 		       sdp->sd_fsname, error);
-		goto fail_gunlock_live;
+		return error;
 	}
 
 	error = gfs_read_sb(sdp, sb_gh.gh_gl, silent);
 	if (error) {
 		printk("GFS: fsid=%s: can't read superblock: %d\n",
 		       sdp->sd_fsname, error);
-		goto fail_gunlock_sb;
+		goto out;
 	}
 
 	/*  Set up the buffer cache and SB for real, now that we know block
@@ -279,13 +312,13 @@ fill_super(struct super_block *sb, void *data, int silent)
 	if (sdp->sd_sb.sb_bsize < bdev_hardsect_size(sb->s_bdev)) {
 		printk("GFS: fsid=%s: FS block size (%u) is too small for device block size (%u)\n",
 		       sdp->sd_fsname, sdp->sd_sb.sb_bsize, bdev_hardsect_size(sb->s_bdev));
-		goto fail_gunlock_sb;
+		goto out;
 	}
 	if (sdp->sd_sb.sb_bsize > PAGE_SIZE) {
 		printk("GFS: fsid=%s: FS block size (%u) is too big for machine page size (%u)\n",
 		       sdp->sd_fsname, sdp->sd_sb.sb_bsize,
 		       (unsigned int)PAGE_SIZE);
-		goto fail_gunlock_sb;
+		goto out;
 	}
 
 	/*  Get rid of buffers from the original block size  */
@@ -293,15 +326,56 @@ fill_super(struct super_block *sb, void *data, int silent)
 	sb_gh.gh_gl->gl_aspace->i_blkbits = sdp->sd_sb.sb_bsize_shift;
 
 	sb_set_blocksize(sb, sdp->sd_sb.sb_bsize);
-	set_blocksize(gfs_diaper_2real(sb->s_bdev), sdp->sd_sb.sb_bsize);
 
-	/*  Read-in journal index inode (but not the file contents, yet)  */
+	/*  Read in the resource index inode  */
 
-	error = gfs_get_jiinode(sdp);
+	error = gfs_get_riinode(sdp);
 	if (error) {
-		printk("GFS: fsid=%s: can't get journal index inode: %d\n",
+		printk("GFS: fsid=%s: can't get resource index inode: %d\n",
 		       sdp->sd_fsname, error);
-		goto fail_gunlock_sb;
+		goto out;
+	}
+
+	/*  Get the root inode  */
+	error = gfs_get_rootinode(sdp);
+	if (error) {
+		printk("GFS: fsid=%s: can't read in root inode: %d\n",
+		       sdp->sd_fsname, error);
+		goto out;
+	}
+	sb->s_root = d_alloc_root(gfs_iget(sdp->sd_rooti, TRUE));
+	if (!sb->s_root) {
+		printk("GFS: can't get root dentry\n");
+		error = -ENOMEM;
+	}
+	sb->s_root->d_op = &gfs_dops;
+
+	/*  Read in the quota inode  */
+	error = gfs_get_qinode(sdp);
+	if (error) {
+		printk("GFS: fsid=%s: can't get quota file inode: %d\n",
+		       sdp->sd_fsname, error);
+		goto out;
+	}
+
+	/*  We're through with the superblock lock  */
+out:
+	gfs_glock_dq_uninit(&sb_gh);
+	super = FALSE;
+	return error;
+}
+
+static int init_journal(struct gfs_sbd *sdp, int undo)
+{
+	struct gfs_holder ji_gh;
+	int error = 0;
+	unsigned int x;
+	int jindex = TRUE;
+	struct task_struct *p;
+
+	if (undo) {
+		jindex = 0;
+		goto fail_recoverd;
 	}
 
 	init_MUTEX(&sdp->sd_jindex_lock);
@@ -312,48 +386,47 @@ fill_super(struct super_block *sb, void *data, int silent)
 	error = gfs_glock_get(sdp, GFS_TRANS_LOCK, &gfs_trans_glops,
 			      CREATE, &sdp->sd_trans_gl);
 	if (error)
-		goto fail_ji_free;
+		goto fail;
 	set_bit(GLF_STICKY, &sdp->sd_trans_gl->gl_flags);
 
-	/*  Upgrade GFS on-disk format version numbers if we need to  */
-
-	if (sdp->sd_args.ar_upgrade) {
-		error = gfs_do_upgrade(sdp, sb_gh.gh_gl);
-		if (error)
-			goto fail_trans_gl;
-	}
-
 	/*  Load in the journal index special file */
 
 	error = gfs_jindex_hold(sdp, &ji_gh);
 	if (error) {
 		printk("GFS: fsid=%s: can't read journal index: %d\n",
 		       sdp->sd_fsname, error);
-		goto fail_trans_gl;
+		goto fail_jindex;
 	}
 
-	/*  Discover this node's journal number (lock module tells us
-	    which one to use), and lock it */
-	error = -EINVAL;
-	if (sdp->sd_lockstruct.ls_jid >= sdp->sd_journals) {
-		printk("GFS: fsid=%s: can't mount journal #%u\n",
-		       sdp->sd_fsname, sdp->sd_lockstruct.ls_jid);
-		printk("GFS: fsid=%s: there are only %u journals (0 - %u)\n",
-		     sdp->sd_fsname, sdp->sd_journals, sdp->sd_journals - 1);
-		goto fail_gunlock_ji;
+	if (sdp->sd_args.ar_spectator) {
+		sdp->sd_jdesc = sdp->sd_jindex[0];
+		sdp->sd_log_seg_free = sdp->sd_jdesc.ji_nsegment;
+		sdp->sd_log_seg_ail2 = 0;
 	}
-	sdp->sd_jdesc = sdp->sd_jindex[sdp->sd_lockstruct.ls_jid];
-	sdp->sd_log_seg_free = sdp->sd_jdesc.ji_nsegment;
-	sdp->sd_log_seg_ail2 = 0;
+	else {
+		/*  Discover this node's journal number (lock module tells us
+		    which one to use), and lock it */
+		error = -EINVAL;
+		if (sdp->sd_lockstruct.ls_jid >= sdp->sd_journals) {
+			printk("GFS: fsid=%s: can't mount journal #%u\n",
+			       sdp->sd_fsname, sdp->sd_lockstruct.ls_jid);
+			printk("GFS: fsid=%s: there are only %u journals (0 - %u)\n",
+			       sdp->sd_fsname, sdp->sd_journals, sdp->sd_journals - 1);
+			goto fail_jindex;
+		}
+		sdp->sd_jdesc = sdp->sd_jindex[sdp->sd_lockstruct.ls_jid];
+		sdp->sd_log_seg_free = sdp->sd_jdesc.ji_nsegment;
+		sdp->sd_log_seg_ail2 = 0;
 
-	error = gfs_glock_nq_num(sdp,
-				 sdp->sd_jdesc.ji_addr, &gfs_meta_glops,
-				 LM_ST_EXCLUSIVE, LM_FLAG_NOEXP,
-				 &sdp->sd_journal_gh);
-	if (error) {
-		printk("GFS: fsid=%s: can't acquire the journal glock: %d\n",
-		       sdp->sd_fsname, error);
-		goto fail_gunlock_ji;
+		error = gfs_glock_nq_num(sdp,
+					 sdp->sd_jdesc.ji_addr, &gfs_meta_glops,
+					 LM_ST_EXCLUSIVE, LM_FLAG_NOEXP,
+					 &sdp->sd_journal_gh);
+		if (error) {
+			printk("GFS: fsid=%s: can't acquire the journal glock: %d\n",
+			       sdp->sd_fsname, error);
+			goto fail_jindex;
+		}
 	}
 
 	if (sdp->sd_lockstruct.ls_first) {
@@ -367,21 +440,19 @@ fill_super(struct super_block *sb, void *data, int silent)
 			if (error) {
 				printk("GFS: fsid=%s: error recovering journal %u: %d\n",
 				       sdp->sd_fsname, x, error);
-				goto fail_gunlock_journal;
+				goto fail_journal_gh;
 			}
 		}
 
 		gfs_lm_others_may_mount(sdp);
-	} else {
+	} else if (!sdp->sd_args.ar_spectator) {
 		/*  We're not the first; replay only our own journal. */
-		error = gfs_recover_journal(sdp,
-					    sdp->sd_lockstruct.ls_jid,
-					    &sdp->sd_jdesc,
-					    TRUE);
+		error = gfs_recover_journal(sdp, sdp->sd_lockstruct.ls_jid,
+									&sdp->sd_jdesc, TRUE);
 		if (error) {
 			printk("GFS: fsid=%s: error recovering my journal: %d\n",
 			       sdp->sd_fsname, error);
-			goto fail_gunlock_journal;
+			goto fail_journal_gh;
 		}
 	}
 
@@ -389,131 +460,76 @@ fill_super(struct super_block *sb, void *data, int silent)
 	jindex = FALSE;
 
 	/*  Disown my Journal glock  */
-
 	sdp->sd_journal_gh.gh_owner = NULL;
 
-	/*  Drop our buffer cache and reread all the things we read before
-	    the journal replay, on the unlikely chance that the replay might
-	    have affected (corrected/updated) the superblock contents
-	    or journal index. */
-
-	error = gfs_read_sb(sdp, sb_gh.gh_gl, FALSE);
-	if (error) {
-		printk("GFS: fsid=%s: can't read superblock: %d\n",
-		       sdp->sd_fsname, error);
-		goto fail_gunlock_journal;
-	}
-
-	gfs_glock_force_drop(sdp->sd_jiinode->i_gl);
-
-	error = gfs_jindex_hold(sdp, &ji_gh);
-	if (error) {
-		printk("GFS: fsid=%s: can't read journal index: %d\n",
-		       sdp->sd_fsname, error);
-		goto fail_gunlock_journal;
-	}
-	gfs_glock_dq_uninit(&ji_gh);
-
 	/*  Make the FS read/write  */
 
 	if (!test_bit(SDF_ROFS, &sdp->sd_flags)) {
 		error = gfs_make_fs_rw(sdp);
 		if (error) {
-			printk("GFS: fsid=%s: can't make FS RW: %d\n",
+			printk("GFS: fsid=%s: can't make file system RW: %d\n",
 			       sdp->sd_fsname, error);
-			goto fail_gunlock_journal;
+			goto fail;
 		}
 	}
 
 	/*  Start up the journal recovery thread  */
 
-	error = kernel_thread(gfs_recoverd, sdp, 0);
-	if (error < 0) {
+	p = kthread_run(gfs_recoverd, sdp, "gfs_recoverd");
+	error = IS_ERR(p);
+	if (error) {
 		printk("GFS: fsid=%s: can't start recoverd thread: %d\n",
 		       sdp->sd_fsname, error);
-		goto fail_make_ro;
+		goto fail_journal_gh;
 	}
-	wait_for_completion(&sdp->sd_thread_completion);
+	sdp->sd_recoverd_process = p;
 
-	/*  Read in the resource index inode  */
+	return 0;
 
-	error = gfs_get_riinode(sdp);
-	if (error) {
-		printk("GFS: fsid=%s: can't get resource index inode: %d\n",
-		       sdp->sd_fsname, error);
-		goto fail_recoverd;
-	}
+fail_recoverd:
+	kthread_stop(sdp->sd_recoverd_process);
+	sdp->sd_recoverd_process = NULL;
 
-	/*  Get the root inode  */
+fail_journal_gh:
+	if (!sdp->sd_args.ar_spectator)
+		gfs_glock_dq_uninit(&sdp->sd_journal_gh);
 
-	error = gfs_get_rootinode(sdp);
-	if (error) {
-		printk("GFS: fsid=%s: can't read in root inode: %d\n",
-		       sdp->sd_fsname, error);
-		goto fail_ri_free;
-	}
+fail_jindex:
+	if (jindex)
+		gfs_glock_dq_uninit(&ji_gh);
+fail:
+	return error;
+}
 
-	/*  Read in the quota inode  */
+static int init_threads(struct gfs_sbd *sdp, int undo)
+{
+	struct task_struct *p;
+	int error = 0;
 
-	error = gfs_get_qinode(sdp);
-	if (error) {
-		printk("GFS: fsid=%s: can't get quota file inode: %d\n",
-		       sdp->sd_fsname, error);
-		goto fail_root_free;
-	}
-
-	/*  Read in the license inode  */
-
-	error = gfs_get_linode(sdp);
-	if (error) {
-		printk("GFS: fsid=%s: can't get license file inode: %d\n",
-		       sdp->sd_fsname, error);
-		goto fail_qi_free;
-	}
-
-	/*  We're through with the superblock lock  */
-
-	gfs_glock_dq_uninit(&sb_gh);
-	super = FALSE;
-
-	/*  Get the root inode/dentry  */
-
-	inode = gfs_iget(sdp->sd_rooti, CREATE);
-	if (!inode) {
-		printk("GFS: fsid=%s: can't get root inode\n", sdp->sd_fsname);
-		error = -ENOMEM;
-		goto fail_li_free;
-	}
-
-	sb->s_root = d_alloc_root(inode);
-	if (!sb->s_root) {
-		iput(inode);
-		printk("GFS: fsid=%s: can't get root dentry\n", sdp->sd_fsname);
-		error = -ENOMEM;
-		goto fail_li_free;
-	}
-
-	/*  Start up the logd thread  */
+	if (undo)
+		goto fail_logd;
 
 	sdp->sd_jindex_refresh_time = jiffies;
 
-	error = kernel_thread(gfs_logd, sdp, 0);
-	if (error < 0) {
+	/*  Start up the logd thread  */
+	p = kthread_run(gfs_logd, sdp, "gfs_logd");
+	error = IS_ERR(p);
+	if (error) {
 		printk("GFS: fsid=%s: can't start logd thread: %d\n",
 		       sdp->sd_fsname, error);
-		goto fail_dput;
+		goto fail;
 	}
-	wait_for_completion(&sdp->sd_thread_completion);
+	sdp->sd_logd_process = p;
 
 	/*  Start up the quotad thread  */
 
-	error = kernel_thread(gfs_quotad, sdp, 0);
+	p = kthread_run(gfs_quotad, sdp, "gfs_quotad");
 	if (error < 0) {
 		printk("GFS: fsid=%s: can't start quotad thread: %d\n",
 		       sdp->sd_fsname, error);
-		goto fail_logd;
+		goto fail_quotad;
 	}
-	wait_for_completion(&sdp->sd_thread_completion);
+	sdp->sd_quotad_process = p;
 
 	/*  Start up the inoded thread  */
 
@@ -521,121 +537,167 @@ fill_super(struct super_block *sb, void *data, int silent)
 	if (error < 0) {
 		printk("GFS: fsid=%s: can't start inoded thread: %d\n",
 		       sdp->sd_fsname, error);
-		goto fail_quotad;
+		goto fail_inoded;
 	}
-	wait_for_completion(&sdp->sd_thread_completion);
+	sdp->sd_inoded_process = p;
+	return 0;
 
+fail_logd:
+	kthread_stop(sdp->sd_logd_process);
+fail_inoded:
+	kthread_stop(sdp->sd_quotad_process);
+fail_quotad:
+	kthread_stop(sdp->sd_logd_process);
+fail:
+	return error;
+}
+
+/**
+ * fill_super - Read in superblock
+ * @sb: The VFS superblock
+ * @data: Mount options
+ * @silent: Don't complain if it's not a GFS filesystem
+ *
+ * Returns: errno
+ *
+ * After cross-linking Linux VFS incore superblock and our GFS incore
+ *   superblock (filesystem instance structures) to one another, we:
+ * -- Init some of our GFS incore superblock, including some temporary
+ *       block-size values (enough to read on-disk superblock).
+ * -- Set up some things in Linux VFS superblock.
+ * -- Mount a lock module, init glock system (incl. glock reclaim daemons),
+ *       and init some important inter-node locks (MOUNT, LIVE, SuperBlock).
+ * -- Read-in the GFS on-disk superblock (1st time, to get enough info
+ *       to do filesystem upgrade and journal replay, incl. journal index).
+ * -- Upgrade on-disk filesystem format (rarely needed).
+ * -- Replay journals (always; replay *all* journals if we're first-to-mount).
+ * -- Read-in on-disk superblock and journal index special file again
+ *       (2nd time, assumed 100% valid now after journal replay).
+ * -- Read-in info on other special (hidden) files (root inode, resource index,
+ *       quota inode, license inode).
+ * -- Start other daemons (journal/log recovery, log tail, quota updates, inode
+ *       reclaim) for periodic maintenance.
+ * 
+ */
+
+static int fill_super(struct super_block *sb, void *data, int silent)
+{
+	struct gfs_sbd *sdp;
+	struct gfs_holder mount_gh;
+	int error;
+
+	sdp = init_sbd(sb);
+	if (!sdp) {
+		printk(KERN_WARNING "GFS: can't alloc struct gfs_sbd\n");
+		return -ENOMEM;
+	}
+
+	error = gfs_make_args((char *)data, &sdp->sd_args, FALSE);
+	if (error) {
+		printk("GFS: can't parse mount arguments\n");
+		goto fail;
+	}
+
+	if (sdp->sd_args.ar_spectator) {
+		sb->s_flags |= MS_RDONLY;
+		set_bit(SDF_ROFS, &sdp->sd_flags);
+	}
+
+	/*  Copy VFS mount flags  */
+
+	if (sb->s_flags & (MS_NOATIME | MS_NODIRATIME))
+		set_bit(SDF_NOATIME, &sdp->sd_flags);
+	if (sb->s_flags & MS_RDONLY)
+		set_bit(SDF_ROFS, &sdp->sd_flags);
+
+	init_vfs(sb, SDF_NOATIME);
+
+	/*  Set up the buffer cache and fill in some fake block size values
+	   to allow us to read-in the on-disk superblock.  */
+
+	sdp->sd_sb.sb_bsize = sb_min_blocksize(sb, GFS_BASIC_BLOCK);
+	sdp->sd_sb.sb_bsize_shift = sb->s_blocksize_bits;
+	sdp->sd_fsb2bb_shift = sdp->sd_sb.sb_bsize_shift - GFS_BASIC_BLOCK_SHIFT;
+	sdp->sd_fsb2bb = 1 << sdp->sd_fsb2bb_shift;
+
+	if (sizeof(struct gfs_sb) > sdp->sd_sb.sb_bsize) {
+		printk("GFS: sizeof(struct gfs_sb) > sdp->sd_sb.sb_bsize\n"
+		       "GFS: %u > %u\n",
+		       (unsigned int)sizeof(struct gfs_sb), sdp->sd_sb.sb_bsize);
+		error = -EINVAL;
+		goto fail;
+	}
+	error = init_names(sdp, silent);
+	if (error)
+		goto fail;
+
+	error = gfs_sys_fs_add(sdp);
+	if (error)
+		goto fail;
+
+	/*  Mount an inter-node lock module, check for local optimizations */
+
+	error = gfs_lm_mount(sdp, silent);
+	if (error)
+		goto fail_sys;
+
+	error = init_locking(sdp, &mount_gh, DO);
+	if (error)
+		goto fail_lm;
+
+	error = init_sb(sdp, silent, DO);
+	if (error)
+		goto fail_locking;
+
+	/*  Read-in journal index inode (but not the file contents, yet)  */
+
+	error = gfs_get_jiinode(sdp);
+	if (error) {
+		printk("GFS: fsid=%s: can't get journal index inode: %d\n",
+		       sdp->sd_fsname, error);
+		goto fail_sb;
+	}
+
+	error = init_journal(sdp, DO);
+	if (error)
+		goto fail_sb;
 	/*  Get a handle on the rename lock  */
 
 	error = gfs_glock_get(sdp, GFS_RENAME_LOCK, &gfs_nondisk_glops,
-			      CREATE, &sdp->sd_rename_gl);
+						  CREATE, &sdp->sd_rename_gl);
 	if (error)
-		goto fail_inoded;
+		goto fail_journal;
+
+	error = init_threads(sdp, DO);
+	if (error)
+		goto fail_journal;
 
 	gfs_proc_fs_add(sdp);
-
 	gfs_glock_dq_uninit(&mount_gh);
 
 	return 0;
 
- fail_inoded:
-	down(&sdp->sd_thread_lock);
-	clear_bit(SDF_INODED_RUN, &sdp->sd_flags);
-	wake_up_process(sdp->sd_inoded_process);
-	up(&sdp->sd_thread_lock);
-	wait_for_completion(&sdp->sd_thread_completion);
+fail_journal:
+	init_journal(sdp, UNDO);
 
- fail_quotad:
-	down(&sdp->sd_thread_lock);
-	clear_bit(SDF_QUOTAD_RUN, &sdp->sd_flags);
-	wake_up_process(sdp->sd_quotad_process);
-	up(&sdp->sd_thread_lock);
-	wait_for_completion(&sdp->sd_thread_completion);
+fail_sb:
+	init_sb(sdp, 0, UNDO);
 
- fail_logd:
-	down(&sdp->sd_thread_lock);
-	clear_bit(SDF_LOGD_RUN, &sdp->sd_flags);
-	wake_up_process(sdp->sd_logd_process);
-	up(&sdp->sd_thread_lock);
-	wait_for_completion(&sdp->sd_thread_completion);
+fail_locking:
+	init_locking(sdp, &mount_gh, UNDO);
 
- fail_dput:
-	dput(sb->s_root);
-
- fail_li_free:
-	gfs_inode_put(sdp->sd_linode);
-
- fail_qi_free:
-	gfs_inode_put(sdp->sd_qinode);
-
- fail_root_free:
-	gfs_inode_put(sdp->sd_rooti);
-
- fail_ri_free:
-	gfs_inode_put(sdp->sd_riinode);
-	gfs_clear_rgrpd(sdp);
-
- fail_recoverd:
-	down(&sdp->sd_thread_lock);
-	clear_bit(SDF_RECOVERD_RUN, &sdp->sd_flags);
-	wake_up_process(sdp->sd_recoverd_process);
-	up(&sdp->sd_thread_lock);
-	wait_for_completion(&sdp->sd_thread_completion);
-
- fail_make_ro:
-	gfs_glock_force_drop(sdp->sd_trans_gl);
-	clear_bit(SDF_JOURNAL_LIVE, &sdp->sd_flags);
-	gfs_unlinked_cleanup(sdp);
-	gfs_quota_cleanup(sdp);
-
- fail_gunlock_journal:
-	gfs_glock_dq_uninit(&sdp->sd_journal_gh);
-
- fail_gunlock_ji:
-	if (jindex)
-		gfs_glock_dq_uninit(&ji_gh);
-
- fail_trans_gl:
-	gfs_glock_put(sdp->sd_trans_gl);
-
- fail_ji_free:
-	gfs_inode_put(sdp->sd_jiinode);
-	gfs_clear_journals(sdp);
-
- fail_gunlock_sb:
-	if (super)
-		gfs_glock_dq_uninit(&sb_gh);
-
- fail_gunlock_live:
-	gfs_glock_dq_uninit(&sdp->sd_live_gh);
-
- fail_gunlock_mount:
-	gfs_glock_dq_uninit(&mount_gh);
-
- fail_glockd:
-	clear_bit(SDF_GLOCKD_RUN, &sdp->sd_flags);
-	wake_up(&sdp->sd_reclaim_wchan);
-	while (sdp->sd_glockd_num--)
-		wait_for_completion(&sdp->sd_thread_completion);
-
-	down(&sdp->sd_thread_lock);
-	clear_bit(SDF_SCAND_RUN, &sdp->sd_flags);
-	wake_up_process(sdp->sd_scand_process);
-	up(&sdp->sd_thread_lock);
-	wait_for_completion(&sdp->sd_thread_completion);
-
- fail_lockproto:
-	gfs_gl_hash_clear(sdp, TRUE);
+fail_lm:
 	gfs_lm_unmount(sdp);
-	gfs_clear_dirty_j(sdp);
 	while (invalidate_inodes(sb))
 		yield();
 
- fail_vfree:
+fail_sys:
+	gfs_sys_fs_del(sdp);
+
+fail:
 	vfree(sdp);
+	sb->s_fs_info = NULL;
 
- fail:
-	set_v2sdp(sb, NULL);
 	return error;
 }
 
@@ -649,7 +711,7 @@ fill_super(struct super_block *sb, void *data, int silent)
 int
 gfs_test_bdev_super(struct super_block *sb, void *data)
 {
-	return (void *)sb->s_bdev == data;	
+	return (void *)sb->s_bdev == data;
 }
 
 /**
@@ -679,60 +741,10 @@ gfs_set_bdev_super(struct super_block *sb, void *data)
  * Returns: the new superblock
  */
 
-struct super_block *
-gfs_get_sb(struct file_system_type *fs_type, int flags,
-	   const char *dev_name, void *data)
+static int gfs_get_sb(struct file_system_type *fs_type, int flags,
+					  const char *dev_name, void *data, struct vfsmount *mnt)
 {
-	struct block_device *real, *diaper;
-	struct super_block *sb;
-	int error = 0;
-
-	real = open_bdev_excl(dev_name, flags, fs_type);
-	if (IS_ERR(real))
-		return (struct super_block *)real;
-
-	diaper = gfs_diaper_get(real, flags);
-	if (IS_ERR(diaper)) {
-		close_bdev_excl(real);
-		return (struct super_block *)diaper;
-	}
-
-	down(&diaper->bd_mount_sem);
-	sb = sget(fs_type, gfs_test_bdev_super, gfs_set_bdev_super, diaper);
-	up(&diaper->bd_mount_sem);
-	if (IS_ERR(sb))
-		goto out;
-
-	if (sb->s_root) {
-		if ((flags ^ sb->s_flags) & MS_RDONLY) {
-			up_write(&sb->s_umount);
-			deactivate_super(sb);
-			sb = ERR_PTR(-EBUSY);
-		}
-		goto out;
-	} else {
-		char buf[BDEVNAME_SIZE];
-
-		sb->s_flags = flags;
-		strlcpy(sb->s_id, bdevname(real, buf), sizeof(sb->s_id));
-		sb->s_old_blocksize = block_size(real);
-		sb_set_blocksize(sb, sb->s_old_blocksize);
-		set_blocksize(real, sb->s_old_blocksize);
-		error = fill_super(sb, data, (flags & MS_VERBOSE) ? 1 : 0);
-		if (error) {
-			up_write(&sb->s_umount);
-			deactivate_super(sb);
-			sb = ERR_PTR(error);
-		} else
-			sb->s_flags |= MS_ACTIVE;
-	}
-
-	return sb;
-
- out:
-	gfs_diaper_put(diaper);
-	close_bdev_excl(real);
-	return sb;
+	return get_sb_bdev(fs_type, flags, dev_name, data, fill_super, mnt);
 }
 
 /**
@@ -746,15 +758,7 @@ gfs_get_sb(struct file_system_type *fs_type, int flags,
 void
 gfs_kill_sb(struct super_block *sb)
 {
-	struct block_device *diaper = sb->s_bdev;
-	struct block_device *real = gfs_diaper_2real(diaper);
-	unsigned long bsize = sb->s_old_blocksize;
-
-	generic_shutdown_super(sb);
-	set_blocksize(diaper, bsize);
-	set_blocksize(real, bsize);
-	gfs_diaper_put(diaper);
-	close_bdev_excl(real);
+	kill_block_super(sb);
 }
 
 struct file_system_type gfs_fs_type = {
