@@ -29,13 +29,12 @@
 #include <linux/in.h>
 #include <linux/buffer_head.h>
 #include <linux/miscdevice.h>
-
-#include <linux/devfs_fs_kernel.h>
+#include <linux/moduleparam.h>
 
 #include <asm/uaccess.h>
 #include <asm/types.h>
 
-#include <linux/gnbd.h>
+#include "gnbd.h"
 
 static int major_nr = 0;
 uint64_t insmod_time;
@@ -69,7 +68,6 @@ static spinlock_t gnbd_lock = SPIN_LOCK_UNLOCKED;
 
 static void gnbd_class_release(struct class_device *class_dev)
 {
-	printk("releasing gnbd class\n");
 	/* FIXME -- What the hell do I have to free up here */
 }
 
@@ -180,9 +178,9 @@ static ssize_t store_sectors(struct class_device *class_dev,
 	set_capacity(dev->disk, size);
 	bdev = bdget_disk(dev->disk, 0);
 	if (bdev) {
-		down(&bdev->bd_inode->i_sem);
+		mutex_lock(&bdev->bd_inode->i_mutex);
 		i_size_write(bdev->bd_inode, (loff_t)size << 9);
-		up(&bdev->bd_inode->i_sem);
+		mutex_unlock(&bdev->bd_inode->i_mutex);
 		bdput(bdev);
 	}
 	up(&dev->do_it_lock);
@@ -276,12 +274,12 @@ static void gnbd_end_request(struct request *req)
 			req, uptodate? "done": "failed");
 
 	if (!uptodate)
-		printk("%s %d called gnbd_end_request with and error\n",
+		printk("%s %d called gnbd_end_request with an error\n",
 		       current->comm, current->pid);	
 	
 	spin_lock_irqsave(q->queue_lock, flags);
 	if (!end_that_request_first(req, uptodate, req->nr_sectors)) {
-		end_that_request_last(req);
+		end_that_request_last(req, uptodate);
 	}
 	spin_unlock_irqrestore(q->queue_lock, flags);
 }
@@ -746,7 +744,7 @@ static int gnbd_ctl_ioctl(struct inode *inode, struct file *file,
 		}
 		error = gnbd_resend_requests(dev, SOCKET_I(inode));
 		if (error){
-			printk("quitting NBD_DO_IT\n");
+			printk("quitting GNBD_DO_IT\n");
 			up(&dev->do_it_lock);
 			fput(file);
 			return error;
@@ -771,7 +769,7 @@ static int gnbd_ctl_ioctl(struct inode *inode, struct file *file,
 		dev->file = NULL;
 		if (file)
 			fput(file);
-		printk("exitting GNBD_DO_IT ioctl\n");
+		printk("exiting GNBD_DO_IT ioctl\n");
 		return error;
 	case GNBD_PING:
 		/* FIXME -- should I allow pings if everything is compeletely
@@ -794,7 +792,7 @@ static int gnbd_ctl_ioctl(struct inode *inode, struct file *file,
 		return 0;
 	case GNBD_GET_TIME:
 		if (copy_to_user((void *)arg, &insmod_time, sizeof(uint64_t))){
-			printk(KERN_WARNING "couldn't compy time argument to user\n");
+			printk(KERN_WARNING "couldn't copy time argument to user\n");
 			return -EFAULT;
 		}
 		return 0;
@@ -845,7 +843,6 @@ static struct miscdevice _gnbd_misc =
 {
         .minor = MISC_DYNAMIC_MINOR,
         .name  = "gnbd_ctl",
-        .devfs_name = "gnbd_ctl",
         .fops = &_gnbd_ctl_fops
 };
 
@@ -926,7 +923,6 @@ static int __init gnbd_init(void)
 	printk(KERN_INFO "gnbd: registered device at major %d\n", major_nr);
 	dprintk(DBG_INIT, "gnbd: debugflags=0x%x\n", debugflags);
 
-	devfs_mk_dir("gnbd_minor");
 	err = class_register(&gnbd_class);
 	if (err)
 		goto out_unregister;
@@ -982,7 +978,6 @@ static int __init gnbd_init(void)
 		disk->fops = &gnbd_fops;
 		disk->private_data = &gnbd_dev[i];
 		sprintf(disk->disk_name, "gnbd%d", i);
-		sprintf(disk->devfs_name, "gnbd_minor/%d", i);
 		set_capacity(disk, 0);
 		add_disk(disk);
 		if(sysfs_create_link(&gnbd_dev[i].class_dev.kobj,
@@ -995,6 +990,7 @@ static int __init gnbd_init(void)
         if (err)
                 goto out_unregister_class;
         
+	do_gettimeofday(&tv); 
 	insmod_time = (uint64_t) tv.tv_sec * 1000000 + tv.tv_usec;
 
 	return 0;
@@ -1036,7 +1032,6 @@ static void __exit gnbd_cleanup(void)
 			kfree(gnbd_dev[i].server_name);
 	}
 	class_unregister(&gnbd_class);
-	devfs_remove("gnbd");
 	unregister_blkdev(major_nr, "gnbd");
 	printk(KERN_INFO "gnbd: unregistered device at major %d\n", major_nr);
 }
@@ -1048,6 +1043,6 @@ MODULE_DESCRIPTION("Network Block Device");
 MODULE_LICENSE("GPL");
 
 #ifndef NDEBUG
-MODULE_PARM(debugflags, "i");
+module_param(debugflags, uint, 0644);
 MODULE_PARM_DESC(debugflags, "flags for controlling debug output");
 #endif
