@@ -26,9 +26,9 @@
 #include <netdb.h>
 #include <arpa/inet.h>
 #include <sys/utsname.h>
-#include <linux/gnbd.h>
 #include <inttypes.h>
 
+#include "gnbd.h"
 #include "fence_return.h"
 #include "gnbd_endian.h"
 #include "gnbd_utils.h"
@@ -81,6 +81,7 @@ int is_clustered = 1;
 #define ACTION_UNFENCE       8
 #define ACTION_LIST_BANNED   9
 #define ACTION_FAIL_SERVER  10
+#define ACTION_GET_UID      11
 
 gnbd_info_t *match_info_name(char *name){
   list_t *item;
@@ -96,12 +97,14 @@ gnbd_info_t *match_info_name(char *name){
 
 
 
-int read_from_server(char *host, uint32_t request, char **buf)
+int talk_to_server(char *host, uint32_t request, char **buf, void *request_data,
+                     ssize_t request_data_size)
 {
   int sock_fd;
-  int n, total;
+  int n, total = 0;
   uint32_t msg;
 
+  *buf = NULL;
   sock_fd = connect_to_server(host, (uint16_t)server_port);
   if (sock_fd < 0){
     printe("cannot connect to server %s (%d) : %s\n", host, sock_fd,
@@ -110,44 +113,55 @@ int read_from_server(char *host, uint32_t request, char **buf)
   }
   msg = cpu_to_be32(request);
   if (write(sock_fd, &msg, sizeof(msg)) != sizeof(msg)){
-    printe("error sending list request to %s : %s\n", host, strerror(errno));
+    printe("error sending request to %s : %s\n", host, strerror(errno));
     exit(1);
   }
+  if (request_data){
+    if (write(sock_fd, request_data, request_data_size) != request_data_size){
+      printe("error sending request data to %s : %s\n", host, strerror(errno));
+      exit(1);
+    }
+  }
   if (read(sock_fd, &msg, sizeof(msg)) != sizeof(msg)){
     printe("error reading reply from %s : %s\n", host, strerror(errno));
     exit(1);
   }
   msg = cpu_to_be32(msg);
   if (msg != EXTERN_SUCCESS_REPLY){
-    printe("list request to %s failed : %s\n", host, strerror(REPLY_ERR(msg)));
+    printe("request to %s failed : %s\n", host, strerror(REPLY_ERR(msg)));
     exit(1);
   }
   if (read(sock_fd, &msg, sizeof(msg)) != sizeof(msg)){
-    printe("error reading size of list from %s : %s\n", host,
+    printe("error reading size of reply from %s : %s\n", host,
            strerror(errno));
     exit(1);
   }
   msg = be32_to_cpu(msg);
+  if (!msg)
+    goto exit;
   *buf = malloc(msg);
   if (*buf == NULL){
-    printe("couldn't allocate memory for list : %s\n", strerror(errno));
+    printe("couldn't allocate memory for server reply : %s\n", strerror(errno));
     exit(1);
   }
   memset(*buf, 0, msg);
-  total = 0;
   while(total < msg){
     n = read(sock_fd, *buf + total, msg - total);
     if (n <= 0){
-      printe("error reading list from server %s : %s\n", host,
+      printe("error reading reply data from server %s : %s\n", host,
              strerror(errno));
       exit(1);
     }
     total += n;
   }
+exit:
   close(sock_fd);
   return total;
 }
 
+#define read_from_server(host, request, buf) \
+  talk_to_server(host, request, buf, NULL, 0)
+
 int start_gnbd_monitor(int minor_nr, int timeout, char *host)
 {
   int ret;
@@ -159,6 +173,10 @@ int start_gnbd_monitor(int minor_nr, int timeout, char *host)
     return 0;
 
   read_from_server(host, EXTERN_NODENAME_REQ, &serv_node);
+  if (!serv_node) {
+    printe("got empty server name\n");
+    return -1;
+  }
 
   snprintf(cmd, 256, "gnbd_monitor %d %d %s", minor_nr, timeout, serv_node);
   ret = system(cmd);
@@ -260,7 +278,8 @@ int usage(void){
 "  -r [GNBD | list] remove specified GNBD(s)\n"
 "  -s <host>        IO fence the host (from known servers by default)\n"
 "  -t <server>      specify a server for the IO fence (with -s or -u)\n"
-"  -u <host>        unfence the host (from known servers by default)\n" 
+"  -u <host>        unfence the host (from known servers by default)\n"
+"  -U <GNBD>        get the Universal Identifier for the specified GNBD\n"
 "  -V               version information\n"
 "  -v               verbose output\n"
 "\n");
@@ -557,6 +576,46 @@ void cleanup_device(char *name, int minor, int fd)
   }
 }
 
+
+#define MULTIPATH_SCRIPT "/etc/dev.d/block/multipath.dev"
+void run_multipath_code(int minor_nr, int add)
+{
+  int i_am_parent, fd;
+  char *envp[6];
+  char devpath[32];
+  char devname[32];
+
+  i_am_parent = daemonize();
+  if (i_am_parent < 0){
+    printm("cannot daemonize to exec multipath code : %s\n", strerror(errno));	
+    return;
+  }
+  if (i_am_parent)
+    return;
+  if ((fd = open("/dev/null", O_RDWR)) < 0) {
+    printm("cannot open /dev/null in daemon : %s\n", strerror(errno));
+    exit(1);
+  }
+  dup2(fd, 0);
+  dup2(fd, 1);
+  dup2(fd, 2);
+  if (fd > 2)
+    close(fd);
+  envp[0] = "HOME=/";
+  envp[1] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";
+  envp[2] = (add)? "ACTION=add" : "ACTION=remove";
+  snprintf(devpath, 31, "DEVPATH=/block/gnbd%d", minor_nr);
+  devpath[31] = '\0';
+  envp[3] = devpath;
+  snprintf(devname, 31, "DEVNAME=/dev/gnbd%d", minor_nr);
+  devname[31] = '\0';
+  envp[4] = devname;
+  envp[5] = NULL;
+  execle(MULTIPATH_SCRIPT, MULTIPATH_SCRIPT, NULL, envp);
+  log_verbose("cannot exec %s : %s", MULTIPATH_SCRIPT, strerror(errno));
+  exit(1);
+}
+
 void remove_gnbd(char *name, int minor, int pid)
 {
   int fd;
@@ -583,6 +642,7 @@ void remove_gnbd(char *name, int minor, int pid)
   if (pid > 0)
   	kill(pid, SIGKILL);
   cleanup_device(name, minor, fd);
+  run_multipath_code(minor, 0);
   close(fd);
 }
 
@@ -968,6 +1028,8 @@ void setclients(char *host)
   int size;
   int minor_nr;
   size = read_from_server(host, EXTERN_NAMES_REQ, &buf);
+  if (!size)
+    return;
   ptr = (import_info_t *)buf;
   while ((char *)ptr < buf + size){
     if (ptr->timeout != 0 && is_clustered == 0){
@@ -980,6 +1042,7 @@ void setclients(char *host)
         exit(1);
       if (start_receiver(minor_nr))
         exit(1);
+      run_multipath_code(minor_nr, 1);
     }
     ptr++;
   }
@@ -991,10 +1054,12 @@ void get_serv_list(char *host)
   import_info_t *ptr;
   int size;
   size = read_from_server(host, EXTERN_NAMES_REQ, &buf);
-  ptr = (import_info_t *)buf;
-  while ((char *)ptr < buf + size){
-    printf("%s\n", ptr->name);
-    ptr++;
+  if (size) {
+    ptr = (import_info_t *)buf;
+    while ((char *)ptr < buf + size){
+      printf("%s\n", ptr->name);
+      ptr++;
+    }
   }
   printf("\n");
   free(buf);
@@ -1006,10 +1071,12 @@ void get_banned_list(char *host)
   node_req_t *ptr;
   int size;
   size = read_from_server(host, EXTERN_LIST_BANNED_REQ, &buf);
-  ptr = (node_req_t *)buf;
-  while ((char *)ptr < buf + size){
-    printf("%s\n", ptr->node_name);
-    ptr++;
+  if (size) {
+    ptr = (node_req_t *)buf;
+    while ((char *)ptr < buf + size){
+      printf("%s\n", ptr->node_name);
+      ptr++;
+    }
   }
   printf("\n");
   free(buf);
@@ -1026,7 +1093,7 @@ void list(void){
     printf("Device name : %s\n"
            "----------------------\n"
            "    Minor # : %d\n"
-           "  Proc name : /dev/gnbd%d\n"
+           " sysfs name : /block/gnbd%d\n"
            "     Server : %s\n"
            "       Port : %d\n"
            "      State : %s %s %s\n"
@@ -1045,6 +1112,34 @@ void list(void){
   }
 }
 
+void get_uid(char *name){
+  list_t *item;
+  gnbd_info_t *gnbd = NULL;
+  char *uid;
+  device_req_t req;
+  int minor_nr = -1;
+
+  sscanf(name, "/block/gnbd%d", &minor_nr);
+  list_foreach(item, &gnbd_list){
+    gnbd = list_entry(item, gnbd_info_t, list);
+    if (minor_nr >= 0 && minor_nr == gnbd->minor_nr)
+      break;
+    if (strncmp(gnbd->name, name, 32) == 0)
+      break;
+    gnbd = NULL;
+  }
+  if (!gnbd){
+    printe("cannot find device '%s'\n", name);
+    exit(1);
+  }
+  strncpy(req.name, gnbd->name, 32);
+  req.name[31] = 0;
+  talk_to_server(gnbd->server_name, EXTERN_UID_REQ, &uid, &req, sizeof(req));
+  if (uid)
+  	printf("%s\n", uid);
+  free(uid);
+}
+  
 void validate_gnbds(void){
   list_t *item;
   char filename[32];
@@ -1079,11 +1174,12 @@ int main(int argc, char **argv)
   int action = 0;
   char *host = NULL;
   char *fence_server = NULL;
+  char *dev = NULL;
   int c;
 
   list_init(&gnbd_list);
   program_name = "gnbd_import";
-  while( (c = getopt(argc, argv, "ac:e:hi:lnOp:qRrs:t:u:Vv")) != -1){
+  while( (c = getopt(argc, argv, "ac:e:hi:lnOp:qRrs:t:U:u:Vv")) != -1){
     switch(c){
     case ':':
     case '?':
@@ -1145,6 +1241,10 @@ int main(int argc, char **argv)
     case 't':
       fence_server = optarg;
       continue;
+    case 'U':
+      set_action(ACTION_GET_UID);
+      dev = optarg;
+      continue;
     case 'u':
       set_action(ACTION_UNFENCE);
       host = optarg;
@@ -1163,7 +1263,7 @@ int main(int argc, char **argv)
       continue;
     default:
       printe("invalid option -- %c\n", c);
-      printf("Please use '-h' for usage.\n");
+      fprintf(stderr, "Please use '-h' for usage.\n");
       exit(1);
     }
   }
@@ -1236,6 +1336,9 @@ int main(int argc, char **argv)
   case ACTION_LIST_BANNED:
     get_banned_list(host);
     return 0;
+  case ACTION_GET_UID:
+    get_uid(dev);
+    return 0;
   default:
     printe("unknown action %d\n", action);
     return 1;
