@@ -41,7 +41,8 @@
 
   What it's _not_:
   - General purpose.  The heap is allocated either at the first call to
-  malloc() or the call to malloc_init(), and is a fixed size.
+  malloc() or the call to malloc_init(), and is a fixed size.  There is no
+  memalign/valloc/pvalloc currently. 
 
   - Super-efficient.  It's reasonably fast in many situations compared to
   glibc's malloc, but records states in the memory blocks to help prevent
@@ -126,7 +127,11 @@
 				   (double free, free after overrun, etc.
 				   for instance) */
 #undef  AGGR_RECLAIM		/* consolidate_all on free (*slow*) */
-#define GDB_HOOK		/* Dump program addresses in malloc_table
+
+#undef  STACKSIZE	/*4	   backtrace to store if DEBUG is set */
+//#define STACKSIZE 4
+
+#undef	GDB_HOOK		/* Dump program addresses in malloc_table
 				   using a fork/exec of gdb (SLOW but fun)
 				   building this defeats the purpose of
 				   a bounded memory allocator, and is only
@@ -143,10 +148,6 @@
 #define ALIGN		(sizeof(void *))
 #define NOBUCKET	((uint16_t)(~0))
 #define MIN_EXTRA	(1<<5)	/* 64 bytes to split a block */
-#define STACKSIZE	4	/* backtrace to store if DEBUG is set
-				   XXX seems to break for values > 5.
-				   unsure why. */
-
 
 /* Misc stuff */
 #define ST_FREE		0xfec3  /* Block is free */
@@ -191,7 +192,9 @@ typedef struct _memblock {
 	uint16_t	mb_bucket;
 	uint16_t	mb_state;
 #ifdef DEBUG
+#ifdef STACKSIZE
 	void		*mb_pc[STACKSIZE];
+#endif
 #endif
 	/* If PARANOID isn't defined, we use the following pointer for
 	   more space. */
@@ -477,7 +480,7 @@ consolidate_all(void)
   without getting stuck.
  */
 void
-zap_mutex(void)
+malloc_zap_mutex(void)
 {
 	pthread_mutex_init(&_alloc_mutex, NULL);
 }
@@ -558,7 +561,7 @@ malloc_init(size_t poolsize)
 		ret = _malloc_init(poolsize);
 		e = errno;
 	}
-	pthread_atfork(NULL, NULL, zap_mutex);
+	pthread_atfork(NULL, NULL, malloc_zap_mutex);
 	pthread_mutex_unlock(&_alloc_mutex);
 	errno = e;
 	return ret;
@@ -713,7 +716,9 @@ malloc(size_t size)
 #endif
 	memblock_t *block;
 #ifdef DEBUG
+#ifdef STACKSIZE
 	int sp;
+#endif
 #endif
 
 	if (size < MIN_SIZE)
@@ -747,12 +752,14 @@ malloc(size_t size)
 	block = search_freestore(size);
 	if (block) {
 #ifdef DEBUG
+#ifdef STACKSIZE
 		for (sp = 0; sp < STACKSIZE; sp++) {
 			assign_address(block->mb_pc, sp);
 			if (!block->mb_pc[sp])
 				break;
 		}
 #endif
+#endif
 #ifdef PARANOID
 		insert_alloc_block(block);
 #endif
@@ -767,12 +774,14 @@ malloc(size_t size)
 	block = search_freestore(size);
 	if (block) {
 #ifdef DEBUG
+#ifdef STACKSIZE
 		for (sp = 0; sp < STACKSIZE; sp++) {
 			assign_address(block->mb_pc, sp);
 			if (!block->mb_pc[sp])
 				break;
 		}
 #endif
+#endif
 #ifdef PARANOID
 		insert_alloc_block(block);
 #endif
@@ -870,10 +879,18 @@ realloc(void *oldp, size_t newsize)
 	memblock_t *oldb;
 #ifdef DEBUG
 	memblock_t *newb;
+#ifdef STACKSIZE
 	int sp;
+#endif
 #endif
 	void *newp;
 
+	if (oldp) {
+		oldb = block(oldp);
+		if (newsize <= oldb->mb_size)
+			return oldp;
+	}
+
 	newp = malloc(newsize);
 
 	if (!newp) {
@@ -888,11 +905,13 @@ realloc(void *oldp, size_t newsize)
 	}
 #ifdef DEBUG
 	newb = block(newp);
+#ifdef STACKSIZE
 	for (sp = 0; sp < STACKSIZE; sp++) {
 		assign_address(newb->mb_pc, sp);
 		if (!newb->mb_pc[sp])
 			break;
 	}
+#endif
 #endif
 	return newp;
 }
@@ -907,7 +926,9 @@ calloc(size_t sz, size_t nmemb)
 	void *p;
 #ifdef DEBUG
 	memblock_t *newb;
+#ifdef STACKSIZE
 	int sp;
+#endif
 #endif
 
 	sz *= nmemb;
@@ -917,11 +938,13 @@ calloc(size_t sz, size_t nmemb)
 
 #ifdef DEBUG
 	newb = block(p);
+#ifdef STACKSIZE
 	for (sp = 0; sp < STACKSIZE; sp++) {
 		assign_address(newb->mb_pc, sp);
 		if (!newb->mb_pc[sp])
 			break;
 	}
+#endif
 #endif
 	memset(p, 0, sz);
 	return p;
@@ -936,13 +959,15 @@ void resolve_stack_gdb(void **, size_t);
   is set.
  */
 void
-malloc_dump_table(void)
+malloc_dump_table(size_t minsize, size_t maxsize)
 {
 #ifdef PARANOID
 	int any = 0;
 	int x;
+#ifdef STACKSIZE
 #ifndef GDB_HOOK
 	int sp;
+#endif
 #endif
 	memblock_t *b;
 
@@ -950,6 +975,10 @@ malloc_dump_table(void)
 	pthread_mutex_lock(&_alloc_mutex);
 	for (x=0; x<BUCKET_COUNT; x++) {
 		for (b = alloc_buckets[x]; b; b = b->mb_next) {
+
+			if (b->mb_size < minsize || b->mb_size > maxsize)
+				continue;
+
 			if (!any)
 				fprintf(stderr,
 					"+++ Memory table dump +++\n");
@@ -961,6 +990,7 @@ malloc_dump_table(void)
 			fprintf(stderr,
 				"  %p (%lu bytes) allocation trace:\n",
 				pointer(b), (unsigned long)b->mb_size);
+#ifdef STACKSIZE
 #ifdef GDB_HOOK
 			resolve_stack_gdb(b->mb_pc, STACKSIZE);
 			fprintf(stderr,"\n");
@@ -968,6 +998,7 @@ malloc_dump_table(void)
 			for (sp = 0; sp < STACKSIZE; sp++)
 				fprintf(stderr,"\t%p\n",b->mb_pc[sp]);
 #endif
+#endif /* STACKSIZE */
 #endif /* DEBUG */
 		}
 	}
@@ -1047,6 +1078,8 @@ malloc_stats(void)
 }
 
 
+#ifdef DEBUG
+#ifdef STACKSIZE
 #ifdef GDB_HOOK
 void
 show_gdb_address(char *buf, size_t buflen, void *address)
@@ -1190,3 +1223,6 @@ resolve_stack_gdb(void **stack, size_t stacksize)
 	exit(0);
 }
 #endif
+#endif
+#endif
+
