@@ -39,11 +39,14 @@ NO=1
 YES_STR="yes"
 INVALIDATEBUFFERS="/bin/true"
 
-logAndPrint()
-{
-	echo $*
-}
+# Grab nfs lock tricks if available
+export NFS_TRICKS=1
+if [ -f "$(dirname $0)/svclib_nfslock" ]; then
+	. $(dirname $0)/svclib_nfslock
+	NFS_TRICKS=0
+fi
 
+. $(dirname $0)/ocf-shellfuncs
 
 meta_data()
 {
@@ -115,6 +118,21 @@ meta_data()
 	    <content type="boolean"/>
         </parameter>
 
+	<!-- 
+        <parameter name="active_monitor">
+            <longdesc lang="en">
+	    	If set, the cluster will spawn an active monitoring 
+		daemon which watches the ability to issue I/Os to the
+		file system.  Requires a file system with O_DIRECT
+		support.
+            </longdesc>
+            <shortdesc lang="en">
+	    	Active Monitoring
+            </shortdesc>
+	    <content type="boolean"/>
+        </parameter>
+	-->
+
 	<parameter name="self_fence">
 	    <longdesc lang="en">
 	        If set and unmounting the file system fails, the node will
@@ -127,6 +145,29 @@ meta_data()
 	    <content type="boolean"/>
 	</parameter>
 
+	<parameter name="nfslock" inherit="service%nfslock">
+	    <longdesc lang="en">
+	        If set and unmounting the file system fails, the node will
+		try to kill lockd and issue reclaims across all remaining
+		network interface cards.
+	    </longdesc>
+	    <shortdesc lang="en">
+	        Enable NFS lock workarounds
+	    </shortdesc>
+	    <content type="boolean"/>
+	</parameter>
+
+	<parameter name="fsid">
+	    <longdesc lang="en">
+	    	File system ID for NFS exports.  This can be overridden
+		in individual nfsclient entries.
+	    </longdesc>
+	    <shortdesc lang="en">
+	    	NFS File system ID
+	    </shortdesc>
+	    <content type="string"/>
+	</parameter>
+
         <parameter name="force_fsck">
             <longdesc lang="en">
                 If set, the file system will be checked (even if
@@ -164,13 +205,15 @@ meta_data()
 	<action name="status" interval="1m" timeout="10"/>
 	<action name="monitor" interval="1m" timeout="10"/>
 
+	<!-- Note: active monitoring is constant and supplants all
+	     check depths -->
 	<!-- Checks to see if we can read from the mountpoint -->
-	<action name="status" depth="10" timeout="30" interval="5m"/>
-	<action name="monitor" depth="10" timeout="30" interval="5m"/>
+	<action name="status" depth="10" timeout="30" interval="30"/>
+	<action name="monitor" depth="10" timeout="30" interval="30"/>
 
 	<!-- Checks to see if we can write to the mountpoint (if !ROFS) -->
-	<action name="status" depth="20" timeout="30" interval="10m"/>
-	<action name="monitor" depth="20" timeout="30" interval="10m"/>
+	<action name="status" depth="20" timeout="30" interval="1m"/>
+	<action name="monitor" depth="20" timeout="30" interval="1m"/>
 
 	<action name="meta-data" timeout="5"/>
 	<action name="verify-all" timeout="5"/>
@@ -178,7 +221,9 @@ meta_data()
 
     <special tag="rgmanager">
 	<attributes maxinstances="1"/>
-        <child type="nfsexport"/>
+        <child type="fs" start="1" stop="3"/>
+        <child type="clusterfs" start="1" stop="3"/>
+        <child type="nfsexport" start="3" stop="1"/>
     </special>
 </resource-agent>
 EOT
@@ -187,31 +232,31 @@ EOT
 verify_name()
 {
 	if [ -z "$OCF_RESKEY_name" ]; then
-		echo "No file system name specified."
-		return 1
+		ocf_log err "No file system name specified."
+		return $OCF_ERR_ARGS
 	fi
-	return 0
+	return $OCF_SUCCESS
 }
 
 
 verify_mountpoint()
 {
 	if [ -z "$OCF_RESKEY_mountpoint" ]; then
-		echo No mount point specified.
-		return 1
+		ocf_log err "No mount point specified."
+		return $OCF_ERR_ARGS
 	fi
 
 	if ! [ -e "$OCF_RESKEY_mountpoint" ]; then
-		echo -n "Mount point $OCF_RESKEY_mountpoint will be created "
-		echo "at mount time."
-		return 0
+		ocf_log info "Mount point $OCF_RESKEY_mountpoint will be "\
+				"created at mount time."
+		return $OCF_SUCCESS
 	fi
 
-	[ -d "$OCF_RESKEY_mountpoint" ] && return 0
+	[ -d "$OCF_RESKEY_mountpoint" ] && return $OCF_SUCCESS
 
-	echo $OCF_RESKEY_mountpoint exists but is not a directory.
+	ocf_log err "$OCF_RESKEY_mountpoint exists but is not a directory."
 	
-	return 1
+	return $OCF_ERR_ARGS
 }
 
 
@@ -220,29 +265,29 @@ real_device()
 	declare dev=$1
 	declare realdev
 
-	[ -z "$dev" ] && return 1
+	[ -z "$dev" ] && return $OCF_ERR_ARGS
 
 	if [ -h "$dev" ]; then 
 		realdev=$(readlink -f $dev)
 		if [ $? -ne 0 ]; then
-			return 1
+			return $OCF_ERR_ARGS
 		fi
 		echo $realdev
-		return 0
+		return $OCF_SUCCESS
 	fi
 
 	if [ -b "$dev" ]; then
 		echo $dev
-	       	return 0
+	       	return $OCF_SUCCESS
 	fi
 		
 	realdev=$(findfs $dev 2> /dev/null)
 	if [ -n "$realdev" ] && [ -b "$realdev" ]; then
 		echo $realdev
-		return 0
+		return $OCF_SUCCESS
 	fi
 
-	return 1
+	return $OCF_ERR_GENERIC
 }
 
 
@@ -251,21 +296,21 @@ verify_device()
 	declare realdev
 
 	if [ -z "$OCF_RESKEY_device" ]; then
-	       echo "No device or label specified."
-	       return 1
+	       ocf_log err "No device or label specified."
+	       return $OCF_ERR_ARGS
 	fi
 
 	realdev=$(real_device $OCF_RESKEY_device)
 	if [ -n "$realdev" ]; then
 		if [ "$realdev" != "$OCF_RESKEY_device" ]; then
-			echo "Specified $OCF_RESKEY_device maps to $realdev"
+			ocf_log info "Specified $OCF_RESKEY_device maps to $realdev"
 		fi
-		return 0
+		return $OCF_SUCCESS
 	fi
 
-	echo "Device or label \"$OCF_RESKEY_device\" not valid"
+	ocf_log err "Device or label \"$OCF_RESKEY_device\" not valid"
 
-	return 1
+	return $OCF_ERR_ARGS
 }
 
 
@@ -280,7 +325,7 @@ verify_fstype()
 		;;
 	*)
 		echo "File system type $OCF_RESKEY_fstype not supported"
-		return 1
+		return $OCF_ERR_ARGS
 		;;
 	esac
 }
@@ -288,7 +333,8 @@ verify_fstype()
 
 verify_options()
 {
-	declare -i ret=0
+	declare -i ret=$OCF_SUCCESS
+	declare o
 
 	#
 	# From mount(8)
@@ -398,7 +444,7 @@ verify_options()
 		esac
 
 		echo Option $o not supported for $OCF_RESKEY_fstype
-		ret=1
+		ret=$OCF_ERR_ARGS
 	done
 
 	return $ret
@@ -407,11 +453,11 @@ verify_options()
 
 verify_all()
 {
-	verify_name || return 1
-	verify_fstype || return 1
-	verify_device || return 1
-	verify_mountpoint || return 1
-	verify_options || return 1
+	verify_name || return $OCF_ERR_ARGS
+	verify_fstype || return $OCF_ERR_ARGS
+	verify_device || return $OCF_ERR_ARGS
+	verify_mountpoint || return $OCF_ERR_ARGS
+	verify_options || return $OCF_ERR_ARGS
 }
 
 
@@ -428,7 +474,7 @@ mountInUse () {
 	typeset junk
 
 	if [ $# -ne 2 ]; then
-		logAndPrint $LOG_ERR "Usage: mountInUse device mount_point".
+		ocf_log err "Usage: mountInUse device mount_point".
 		return $FAIL
 	fi
 
@@ -461,14 +507,14 @@ isMounted () {
 	typeset dev tmp_dev
 
 	if [ $# -ne 2 ]; then
-		logAndPrint $LOG_ERR "Usage: isMounted device mount_point"
+		ocf_log err "Usage: isMounted device mount_point"
 		return $FAIL
 	fi
 
 	dev=$(real_device $1)
 	if [ -z "$dev" ]; then
-		logAndPrint $LOG_ERR \
-			"isMounted: Could not match $1 with a real device"
+		ocf_log err \
+			"fs (isMounted): Could not match $1 with a real device"
 		return $FAIL
 	fi
 	mp=$2
@@ -484,7 +530,7 @@ isMounted () {
 			# place
 			#
 			if [ -n "$tmp_mp"  -a "$tmp_mp"  != "$mp" ]; then
-				logAndPrint $LOG_WARNING "\
+				ocf_log warn "\
 Device $dev is mounted on $tmp_mp instead of $mp"
 			fi
 			return $YES
@@ -495,6 +541,65 @@ Device $dev is mounted on $tmp_mp instead of $mp"
 }
 
 
+# 
+# isAlive mount_point
+# 
+# Check to see if mount_point is alive (testing read/write)
+# 
+isAlive()
+{
+	declare mount_point
+	declare file=".writable_test"
+	declare rw
+	
+	if [ $# -ne 1 ]; then
+	        ocf_log err "Usage: isAlive mount_point"
+		return $FAIL
+	fi
+	mount_point=$1
+	
+	test -d $mount_point
+	if [ $? -ne 0 ]; then
+		ocf_log err "fs (isAlive): $mount_point is not a directory"
+		return $FAIL
+	fi
+	
+	[ $OCF_CHECK_LEVEL -lt 10 ] && return $YES
+	
+	# depth 10 test (read test)
+	ls $mount_point > /dev/null 2> /dev/null
+	if [ $? -ne 0 ]; then
+	       return $NO
+	fi
+	
+	[ $OCF_CHECK_LEVEL -lt 20 ] && return $YES
+	
+	# depth 20 check (write test)
+	rw=$YES
+	for o in `echo $OCF_RESKEY_options | sed -e s/,/\ /g`; do
+                if [ "$o" = "ro" ]; then
+		        rw=$NO
+                fi
+	done
+	if [ $rw -eq $YES ]; then
+	        file=$mount_point/$file
+		while true; do
+			if [ -e "$file" ]; then
+				file=${file}_tmp
+				continue
+			else
+			        break
+			fi
+		done
+		touch $file > /dev/null 2> /dev/null
+		[ $? -ne 0 ] && return $NO
+		rm -f $file > /dev/null 2> /dev/null
+	fi
+	
+	return $YES
+}
+
+
 #
 # killMountProcesses mount_point
 #
@@ -509,14 +614,14 @@ killMountProcesses()
 	typeset try
 
 	if [ $# -ne 1 ]; then
-		logAndPrint $LOG_ERR \
+		ocf_log err \
 			"Usage: killMountProcesses mount_point"
 		return $FAIL
 	fi
 
 	typeset mp=$1
 
-	logAndPrint $LOG_INFO "Forcefully unmounting $mp"
+	ocf_log notice "Forcefully unmounting $mp"
 
 	#
 	# Not all distributions have lsof.  If not use fuser.  If it
@@ -533,7 +638,7 @@ killMountProcesses()
 	fi             
 
 	if [ -z "$have_lsof" -a -z "$have_fuser" ]; then
-		logAndPrint $LOG_WARNING \
+		ocf_log warn \
 	"Cannot forcefully unmount $mp; cannot find lsof or fuser commands"
 		return $FAIL
 	fi
@@ -550,10 +655,10 @@ killMountProcesses()
 				fi
 
 				if [ $try -eq 1 ]; then
-					logAndPrint $LOG_WARNING \
+					ocf_log warn \
 			 	  "killing process $pid ($user $command $mp)"
 				elif [ $try -eq 3 ]; then
-					logAndPrint $LOG_CRIT \
+					ocf_log crit \
 		    		  "Could not clean up mountpoint $mp"
 				ret=$FAIL
 				fi
@@ -563,7 +668,7 @@ killMountProcesses()
 				else
 					kill -TERM $pid
 				fi
-			done < <(lsof -b 2>/dev/null | \
+			done < <(lsof -bn 2>/dev/null | \
 			    grep -E "$mp(/.*|)\$" | \
 			    awk '{print $1,$2,$3}' | \
 			    sort -u -k 1,3)
@@ -578,10 +683,10 @@ killMountProcesses()
 				fi
 
 				if [ $try -eq 1 ]; then
-					logAndPrint $LOG_WARNING \
+					ocf_log warn \
 			 	  "killing process $pid ($user $command $mp)"
 				elif [ $try -eq 3 ]; then
-					logAndPrint $LOG_CRIT \
+					ocf_log crit \
 				    "Could not clean up mount point $mp"
 					ret=$FAIL
 				fi
@@ -602,6 +707,139 @@ killMountProcesses()
 	return $ret
 }
 
+
+activeMonitor() {
+	declare monpath=$OCF_RESKEY_mountpoint/.clumanager
+	declare p
+	declare pid
+
+	if [ -z "$OCF_RESKEY_mountpoint" ]; then
+		ocf_log err "activeMonitor: No mount point specified"
+		return $OCF_ERR_ARGS
+	fi
+
+	if [ "$OCF_RESKEY_active_monitor" != "1" ] &&
+	   [ "$OCF_RESKEY_active_monitor" != "yes" ]; then
+		# Nothing bad happened; but no active monitoring specified.
+		return $OCF_SUCCESS
+	fi
+
+	if [ "$OCF_RESKEY_self_fence" = "1" ] ||
+	   [ "$OCF_RESKEY_self_fence" = "yes" ]; then
+		args="-i 2 -a reboot"
+	else
+		args="-i 2"
+	fi
+
+	case $1 in
+	start)
+		ocf_log info "Starting active monitoring of $OCF_RESKEY_mountpoint"
+		mkdir -p $(dirname $monpath) || return $OCF_ERR_GENERIC
+		devmon $args -p $monpath/devmon.data -P $monpath/devmon.pid
+		;;
+	stop)
+		ocf_log info "Stopping active monitoring of $OCF_RESKEY_mountpoint"
+		if ! [ -f $monpath/devmon.pid ]; then
+			# Someone removed the file or it wasn't there for
+			# some reason... Force unmount will kill us
+			return 0
+		fi
+
+		pid=$(cat $monpath/devmon.pid)
+		if [ -z "$pid" ]; then
+			# Someone emptied the file?
+			return 0
+		fi
+
+		for p in $(pidof devmon); do
+			if [ "$pid" = "$p" ]; then
+				ocf_log debug "Killing devmon $p for $OCF_RESKEY_mountpoint"
+				kill -TERM $p
+				return 0
+			fi
+		done
+		# none matching
+
+		return 0
+		;;
+	status)
+		pid=$(cat $monpath/devmon.pid)
+		for p in $(pidof devmon); do
+			if [ "$pid" = "$p" ]; then
+				return 0
+			fi
+		done
+
+		# none matching
+		ocf_log err "Active Monitor for $OCF_RESKEY_mountpoint has exited"
+		return $OCF_ERR_GENERIC
+		;;
+	*)
+		ocf_log err "usage: activeMonitor <start|stop|status>"
+		return $OCF_ERR_ARGS
+		;;
+	esac
+}
+
+
+#
+# Enable quotas on the mount point if the user requested them
+#
+enable_fs_quotas()
+{
+	declare -i need_check=0
+	declare quotaopts=""
+	declare mopt
+	declare opts=$1
+	declare mp=$2
+
+	if [ -z "`which quotaon`" ]; then
+		ocf_log err "quotaon not found in $PATH"
+		return $OCF_ERR_GENERIC
+	fi
+
+	for mopt in `echo $opts | sed -e s/,/\ /g`; do
+		case $mopt in
+		usrquota)
+			quotaopts="u$quotaopts"
+			continue
+			;;
+		grpquota)
+			quotaopts="g$quotaopts"
+			continue
+			;;
+		noquota)
+			quotaopts=""
+			return 0
+			;;
+		esac
+	done
+
+	[ -z "$quotaopts" ] && return 0
+
+	# Ok, create quota files if they don't exist
+	for f in quota.user aquota.user quota.group aquota.group; do
+		if ! [ -f "$mp/$f" ]; then
+			ocf_log info "$mp/$f was missing - creating"
+			touch "$mp/$f" 
+			chmod 600 "$mp/$f"
+			need_check=1
+		fi
+	done
+
+	if [ $need_check -eq 1 ]; then
+		ocf_log info "Checking quota info in $mp"
+		quotacheck -$quotaopts $mp
+	fi
+
+	ocf_log info "Enabling Quotas on $mp"
+	ocf_log debug "quotaon -$quotaopts $mp"
+	quotaon -$quotaopts $mp
+
+	return $?
+}
+
+
 #
 # startFilesystem
 #
@@ -625,7 +863,7 @@ startFilesystem() {
 	/*)			# found it
 	  	;;
 	*)	 		# invalid format
-			logAndPrint $LOG_ERR \
+			ocf_log err \
 "startFilesystem: Invalid mount point format (must begin with a '/'): \'$mp\'"
 	    	return $FAIL
 	    	;;
@@ -636,7 +874,7 @@ startFilesystem() {
 	#
 	dev=$(real_device $OCF_RESKEY_device)
 	if [ -z "$dev" ]; then
-			logAndPrint $LOG_ERR "\
+			ocf_log err "\
 startFilesystem: Could not match $OCF_RESKEY_device with a real device"
 			return $FAIL
 	fi
@@ -646,12 +884,12 @@ startFilesystem: Could not match $OCF_RESKEY_device with a real device"
 	#
 	if [ -e "$mp" ]; then
 		if ! [ -d "$mp" ]; then
-			logAndPrint $LOG_ERR "\
+			ocf_log err"\
 startFilesystem: Mount point $mp exists but is not a directory"
 			return $FAIL
 		fi
 	else
-		logAndPrint $LOG_INFO "\
+		ocf_log err "\
 startFilesystem: Creating mount point $mp for device $dev"
 		mkdir -p $mp
 	fi
@@ -676,7 +914,7 @@ startFilesystem: Creating mount point $mp for device $dev"
 	isMounted $dev $mp
 	case $? in
 	$YES)		# already mounted
-		logAndPrint $LOG_DEBUG "$dev already mounted"
+		ocf_log debug "$dev already mounted"
 		return $SUCCESS
 		;;
 	$NO)		# not mounted, continue
@@ -696,7 +934,7 @@ startFilesystem: Creating mount point $mp for device $dev"
 	mountInUse $dev $mp
 	case $? in
 	$YES)		# uh oh, someone is using the device or mount point
-		logAndPrint $LOG_ERR "\
+		ocf_log err "\
 Cannot mount $dev on $mp, the device or mount point is already in use!"
 		return $FAIL
 		;;
@@ -706,7 +944,7 @@ Cannot mount $dev on $mp, the device or mount point is already in use!"
 		return $FAIL
 		;;
 	*)
-		logAndPrint $LOG_ERR "Unknown return from mountInUse"
+		ocf_log err "Unknown return from mountInUse"
 		return $FAIL
 		;;
 	esac
@@ -719,7 +957,7 @@ Cannot mount $dev on $mp, the device or mount point is already in use!"
 		mkdir -p $mp			# create the mount point
 		ret_val=$?
 		if [ $ret_val -ne 0 ]; then
-			logAndPrint $LOG_ERR \
+			ocf_log err \
 				"'mkdir -p $mp' failed, error=$ret_val"
 			return $FAIL
 		fi
@@ -759,7 +997,7 @@ Cannot mount $dev on $mp, the device or mount point is already in use!"
 	"")       typeset fsck_needed=yes ;;		# assume fsck
 	*)
 		typeset fsck_needed=yes 		# assume fsck
-	     	logAndPrint $LOG_WARNING "\
+	     	ocf_log warn "\
 Unknown file system type '$fstype' for device $dev.  Assuming fsck is required."
 		;;
 	esac
@@ -771,13 +1009,13 @@ Unknown file system type '$fstype' for device $dev.  Assuming fsck is required."
 	if [ -n "$fsck_needed" ] || [ "${OCF_RESKEY_force_fsck}" = "yes" ] ||\
 	   [ "${OCF_RESKEY_force_fsck}" = "1" ]; then
 		typeset fsck_log=/tmp/$(basename $dev).fsck.log
-		logAndPrint $LOG_DEBUG "Running fsck on $dev"
+		ocf_log debug "Running fsck on $dev"
 		fsck -p $dev >> $fsck_log 2>&1
 		ret_val=$?
 		if [ $ret_val -gt 1 ]; then
-			logAndPrint $LOG_ERR "\
+			ocf_log err "\
 'fsck -p $dev' failed, error=$ret_val; check $fsck_log for errors"
-			logAndPrint $LOG_DEBUG "Invalidating buffers for $dev"
+			ocf_log debug "Invalidating buffers for $dev"
 			$INVALIDATEBUFFERS -f $dev
 			return $FAIL
 		fi
@@ -787,14 +1025,29 @@ Unknown file system type '$fstype' for device $dev.  Assuming fsck is required."
 	#
 	# Mount the device
 	#
-	logAndPrint $LOG_DEBUG "mount $fstype_option $mount_options $dev $mp"
+	ocf_log info "mounting $dev on $mp"
+	ocf_log debug "mount $fstype_option $mount_options $dev $mp"
 	mount $fstype_option $mount_options $dev $mp
 	ret_val=$?
 	if [ $ret_val -ne 0 ]; then
-		logAndPrint $LOG_ERR "\
+		ocf_log err "\
 'mount $fstype_option $mount_options $dev $mp' failed, error=$ret_val"
 		return $FAIL
 	fi
+
+	#
+	# Create this for the NFS NLM broadcast bit
+	#
+	if [ $NFS_TRICKS -eq 0 ]; then
+		if [ "$OCF_RESKEY_nfslock" = "yes" ] || \
+	   	   [ "$OCF_RESKEY_nfslock" = "1" ]; then
+			mkdir -p $mp/.clumanager/statd
+			notify_list_merge $mp/.clumanager/statd
+		fi
+	fi
+
+	enable_fs_quotas $opts $mp
+	activeMonitor start || return $OCF_ERR_GENERIC
 	
 	return $SUCCESS
 }
@@ -828,18 +1081,19 @@ stopFilesystem() {
 	/*)			# found it
 	  	;;
 	*)	 		# invalid format
-			logAndPrint $LOG_ERR \
+			ocf_log err \
 "startFilesystem: Invalid mount point format (must begin with a '/'): \'$mp\'"
 	    	return $FAIL
 	    	;;
 	esac
 	
+
 	#
 	# Get the device
 	#
 	dev=$(real_device $OCF_RESKEY_device)
 	if [ -z "$dev" ]; then
-			logAndPrint $LOG_ERR "\
+			ocf_log err "\
 stop: Could not match $OCF_RESKEY_device with a real device"
 			return $FAIL
 	fi
@@ -863,16 +1117,14 @@ stop: Could not match $OCF_RESKEY_device with a real device"
 		esac
 	fi
 
-
 	#
 	# Unmount the device.  
 	#
-
 	while [ ! "$done" ]; do
 		isMounted $dev $mp
 		case $? in
 		$NO)
-			logAndPrint $LOG_INFO "$dev is not mounted"
+			ocf_log info "$dev is not mounted"
 			umount_failed=
 			done=$YES
 			;;
@@ -881,8 +1133,11 @@ stop: Could not match $OCF_RESKEY_device with a real device"
 			;;
 		$YES)
 			sync; sync; sync
-			logAndPrint $LOG_INFO "unmounting $mp"
+			ocf_log info "unmounting $mp"
 
+			activeMonitor stop || return $OCF_ERR_GENERIC
+
+			quotaoff -gu $mp &> /dev/null
 			umount $mp
 			if  [ $? -eq 0 ]; then
 				umount_failed=
@@ -894,6 +1149,21 @@ stop: Could not match $OCF_RESKEY_device with a real device"
 
 			if [ "$force_umount" ]; then
 				killMountProcesses $mp
+				if [ $try -eq 1 ]; then
+	        		  if [ "$OCF_RESKEY_nfslock" = "yes" ] || \
+				     [ "$OCF_RESKEY_nfslock" = "1" ]; then
+				    ocf_log warning \
+					"Dropping node-wide NFS locks"
+	          		    mkdir -p $mp/.clumanager/statd
+				    # Copy out the notify list; our 
+				    # IPs are already torn down
+				    if notify_list_store $mp/.clumanager/statd
+				    then
+				      notify_list_broadcast \
+				        $mp/.clumanager/statd
+				    fi
+				  fi
+				fi
 			fi
 
 			if [ $try -ge $max_tries ]; then
@@ -917,10 +1187,10 @@ stop: Could not match $OCF_RESKEY_device with a real device"
 	done # while 
 
 	if [ -n "$umount_failed" ]; then
-		logAndPrint $LOG_ERR "'umount $mp' failed, error=$ret_val"
+		ocf_log err "'umount $mp' failed, error=$ret_val"
 
 		if [ "$self_fence" ]; then
-			logAndPrint $LOG_ALERT "umount failed - REBOOTING"
+			ocf_log alert "umount failed - REBOOTING"
 			sync
 			reboot -fn
 		fi
@@ -940,19 +1210,37 @@ stop)
 	stopFilesystem
 	exit $?
 	;;
-status)
-	isMounted ${OCF_RESKEY_device} ${OCF_RESKEY_mountpoint}
-	exit $?
+status|monitor)
+  	isMounted ${OCF_RESKEY_device} ${OCF_RESKEY_mountpoint}
+ 	if [ $? -ne $YES ]; then
+		ocf_log err "fs:${OCF_RESKEY_name}: ${OCF_RESKEY_device} is not mounted on ${OCF_RESKEY_mountpoint}"
+		exit $OCF_ERR_GENERIC
+	fi
+
+	if [ "$OCF_RESKEY_active_monitor" = "yes" ] ||
+	   [ "$OCF_RESKEY_active_monitor" = "1" ]; then
+
+	   	activeMonitor status
+		[ $? -eq 0 ] && exit 0
+		ocf_log err "fs:${OCF_RESKEY_name}: Active Monitoring reported a failure"
+		exit $OCF_ERR_GENERIC
+	fi
+ 	
+ 	isAlive ${OCF_RESKEY_mountpoint}
+ 	[ $? -eq $YES ] && exit 0
+
+	ocf_log err "fs:${OCF_RESKEY_name}: Mount point is not accessible!"
+	exit $OCF_ERR_GENERIC
 	;;
 restart)
 	stopFilesystem
 	if [ $? -ne 0 ]; then
-		exit 1
+		exit $OCF_ERR_GENERIC
 	fi
 
 	startFilesystem
 	if [ $? -ne 0 ]; then
-		exit 1
+		exit $OCF_ERR_GENERIC
 	fi
 
 	exit 0
@@ -965,6 +1253,10 @@ verify-all)
 	verify_all
 	exit $?
 	;;
+*)
+	echo "usage: $0 {start|stop|status|monitor|restart|meta-data|verify-all}"
+	exit $OCF_ERR_GENERIC
+	;;
 esac
 
 exit 0
